'''
Created on 17 Dec 2023

@author: jacklok
'''

from datetime import datetime
from trexmodel import program_conf
from trexprogram.reward_program.reward_program_base import EntitledVoucherSummary
from trexmodel.models.datastore.voucher_models import MerchantVoucher
from trexmodel.models.datastore.reward_models import CustomerEntitledVoucher
from trexlib.utils.string_util import is_not_empty
from trexmodel.models.datastore.model_decorators import model_transactional
from trexmodel.utils.model.model_util import generate_transaction_id
from trexmodel.models.datastore.redeem_models import RedemptionCatalogueTransaction,\
    CustomerRedemption
from trexanalytics.bigquery_upstream_data_config import create_partnership_transaction_upstream_for_merchant
from trexmodel.models.datastore.customer_model_helpers import update_customer_entiteld_voucher_summary_with_customer_new_voucher
from trexmodel.models.datastore.message_model_helper import create_redeem_catalogue_item_message
from trexprogram.utils.reward_program_helper import calculate_effective_date,\
    calculate_expiry_date
from trexconf.program_conf import REWARD_PROGRAM_DATE_FORMAT
from google.cloud import ndb
from trexmodel.models.datastore.merchant_models import Outlet
from trexmodel.models.datastore.customer_models import Customer
from trexmodel.models.datastore.partnership_models import PartnershipRewardTransaction
from trexlib.libs.google.gae_app_logger import AppLogger

logger = AppLogger.get_logger('helper')

@model_transactional(desc='giveaway_redeem_catalogue_item')
def giveaway_redeem_catalogue_item(customer, redeem_item_details, redeem_reward_format, 
                                   redemption_catalogue_key, voucher_key, transaction_id=None, 
                                   is_partnership_redemption=False, partner_merchant_acct=None):
    redeemed_datetime       = datetime.utcnow()
    redeem_reward_amount    = redeem_item_details.get('redeem_reward_amount')
    
    if redeem_reward_format in (program_conf.REWARD_FORMAT_POINT,program_conf.REWARD_FORMAT_STAMP) :
        reward_summary      = customer.reward_summary
        
        logger.debug('********************************')
        logger.debug('customer reward_summary=%s', reward_summary)
        logger.debug('********************************')
        
        
        
        if is_partnership_redemption==False and reward_summary.get(redeem_reward_format).get('amount') < redeem_reward_amount:
            raise Exception('Not sufficient reward amount to redeem')
    
        else:
            if transaction_id is None:
                transaction_id = generate_transaction_id(prefix='r')
                
            redemption_catalogue_transction_summary = {
                                                        'redemption_catalogue_key'  : redemption_catalogue_key,
                                                        'voucher_key'               : voucher_key,
                                                        
                                                        }
            
            if is_partnership_redemption==False:
                CustomerRedemption.create(customer, 
                                    reward_format                           = redeem_reward_format,
                                    redeemed_amount                         = redeem_reward_amount,
                                    redeemed_datetime                       = redeemed_datetime, 
                                    transaction_id                          = transaction_id,
                                    redemption_catalogue_transction_summary = redemption_catalogue_transction_summary,
                                    is_partnership_redemption               = is_partnership_redemption,
                                    )
            
            redemption_catalogue_transaction = RedemptionCatalogueTransaction.create(
                                                  ndb.Key(urlsafe=redemption_catalogue_key).get(),
                                                  voucher_key, 
                                                  customer,
                                                  transaction_id,
                                                  redeemed_datetime,
                                                  reward_format=redeem_reward_format,
                                                  redeem_reward_amount=redeem_reward_amount,
                                                  )
            
            reward_summary = __giveaway_voucher_from_redemption_catalogue_item(customer, redeem_item_details, transaction_id, redeemed_datetime, partner_merchant_acct=partner_merchant_acct)
            
            create_redeem_catalogue_item_message(customer, reward_summary.entitled_voucher_summary, redemption_catalogue_transaction)
            
            return reward_summary
                


def __giveaway_voucher_from_redemption_catalogue_item(customer, redeem_item_details, transaction_id, transact_datetime, partner_merchant_acct=None):
    logger.debug('---__giveaway_voucher_from_redemption_catalogue_item---')
    
    voucher_key             = redeem_item_details.get('voucher_key')
    effective_type          = redeem_item_details.get('effective_type')
    effective_value         = redeem_item_details.get('effective_value')
    effective_date_str      = redeem_item_details.get('effective_date')
    
    if effective_type == program_conf.REWARD_EFFECTIVE_TYPE_SPECIFIC_DATE:
        if is_not_empty(effective_date_str):
            effective_date = datetime.strptime(effective_date_str, REWARD_PROGRAM_DATE_FORMAT)
    else:
        effective_date = calculate_effective_date(effective_type, effective_value, start_date = transact_datetime)
    
    expiration_type         = redeem_item_details.get('expiration_type')
    expiration_value        = redeem_item_details.get('expiration_value')
     
    expiry_date             = calculate_expiry_date(expiration_type, expiration_value, start_date=effective_date)
    
    voucher_amount          = redeem_item_details.get('voucher_amount')
    
    merchant_voucher = MerchantVoucher.fetch(voucher_key)
    customer_entitled_voucher_list = []
    reward_summary = EntitledVoucherSummary(transaction_id=transaction_id)
    
    logger.debug('merchant_voucher=%s', merchant_voucher)
    logger.debug('voucher_amount=%s', voucher_amount)
    
    if merchant_voucher:
        entitled_voucher_summary = customer.entitled_voucher_summary or {}
        
        logger.debug('entitled_voucher_summary=%s', entitled_voucher_summary)
        
        for v in range(voucher_amount):
            customer_entitled_voucher = CustomerEntitledVoucher.create(
                                                            merchant_voucher,
                                                            customer, 
                                                            transaction_id          = transaction_id,
                                                            rewarded_datetime       = transact_datetime,
                                                            effective_date          = effective_date,
                                                            expiry_date             = expiry_date,
                                                            partner_merchant_acct   = partner_merchant_acct,
                                                            )
            
            logger.debug('customer_entitled_voucher=%s', customer_entitled_voucher)
            
            update_customer_entiteld_voucher_summary_with_customer_new_voucher(entitled_voucher_summary, customer_entitled_voucher)
            customer_entitled_voucher_list.append(customer_entitled_voucher)
        
        customer.entitled_voucher_summary = entitled_voucher_summary    
        customer.put()
        
        reward_summary.add(merchant_voucher, 
                                   customer_entitled_voucher_list) 
        
    return reward_summary

def __get_redemption_voucher_label(partner_merchant_acct, voucher_key):
    published_voucher_configuration = partner_merchant_acct.published_voucher_configuration
    for voucher_details in published_voucher_configuration.get('vouchers'):
        if voucher_details.get('voucher_key') == voucher_key:
            return voucher_details.get('label')
    return ''


@model_transactional(desc='_giveaway_redemption_catalogue_item_to_partner_customer_and_transfer_reward_to_partner_customer')
def giveaway_redemption_catalogue_item_to_partner_customer_and_transfer_reward_to_partner_customer(
        merchant_acct, partner_merchant_acct, merchant_customer, redeem_item_details, redeem_reward_format, redemption_catalogue_key, voucher_key,
        ):
    hq_outlet       = Outlet.get_head_quarter_outlet(partner_merchant_acct)
    if hq_outlet is None:
        raise Exception('The merchant do not have head quarter outlet')
    else:
        logger.info('hq_outlet=%s', hq_outlet)
        
        if redeem_item_details:
            user_acct               = merchant_customer.registered_user_acct
            transact_datetime       = datetime.utcnow()
            redeem_reward_amount    = redeem_item_details.get('redeem_reward_amount')
            
            logger.info('redeem_reward_amount=%s', redeem_reward_amount)
            
            partner_customer = Customer.get_by_reference_code(user_acct.reference_code, partner_merchant_acct)
            if partner_customer is None:
                partner_customer = Customer.create_from_user(user_acct, outlet=hq_outlet, merchant_acct=partner_merchant_acct)
                logger.info('Partner customer account have been created')
            
            if partner_customer is not None:
                
                logger.info('Partner customer_reward_summary=%s', partner_customer.reward_summary)
                logger.info('Merchant customer_reward_summary=%s', merchant_customer.reward_summary)
                
                voucher_label = __get_redemption_voucher_label(partner_merchant_acct, voucher_key)
                customer_redemption = CustomerRedemption.create(merchant_customer, 
                                        reward_format                           = redeem_reward_format,
                                        redeemed_amount                         = redeem_reward_amount,
                                        redeemed_datetime                       = transact_datetime, 
                                        is_partnership_redemption               = True,
                                        is_allow_to_revert                      = True,
                                        remarks                                 = 'This partnership point redemption, where point have been transferred to %s for %s entitlement' % (partner_merchant_acct.brand_name, voucher_label),
                                    )
                
                transaction_id = customer_redemption.transaction_id
                
                logger.info('transaction_id=%s', transaction_id)
                
                reward_summary = giveaway_redeem_catalogue_item(partner_customer, redeem_item_details, redeem_reward_format, 
                                                       redemption_catalogue_key, voucher_key,
                                                       is_partnership_redemption    = True,
                                                       transaction_id               = transaction_id,
                                                       partner_merchant_acct        = merchant_acct,
                                                       )
                
                logger.info('Partner customer redeemed item have been giveaway')
                
                
                
                
                
                partnership_reward_transaction = PartnershipRewardTransaction.create(merchant_acct, partner_merchant_acct, 
                                                    user_acct, redeem_reward_amount,
                                                    reward_summary  = reward_summary,
                                                    )
                
                logger.info('Partnership reward transaction have been created')
                
                create_partnership_transaction_upstream_for_merchant(partnership_reward_transaction)
                
                return merchant_customer    
        else:
            raise Exception('Redeem item is not found')


'''
Created on 30 Jun 2021

@author: jacklok
'''

from flask import Blueprint, session 
from flask_restful import Resource, abort
import logging
from flask_httpauth import HTTPBasicAuth
from flask_restful import Api
from trexmodel.utils.model.model_util import create_db_client
from trexmodel.models.datastore.merchant_models import MerchantUser
import hashlib
from trexlib.utils.string_util import random_string, is_not_empty
from datetime import datetime, timedelta
from trexconf import conf as api_conf
from trexlib.utils.crypto_util import encrypt_json
from trexlib.libs.flask_wtf.request_wrapper import session_value, request_values,\
    request_headers
from trexmodel.models.datastore.loyalty_models import LoyaltyDeviceSetting
from trexapi.utils.push_notification_helper import create_push_notification
from trexweb.libs.http import create_rest_message
from trexapi.utils.api_helpers import StatusCode
from trexconf.conf import DEFAULT_LANGUAGE

logger = logging.getLogger('api')
#logger = logging.getLogger('debug')



device_util_bp = Blueprint('device_util_bp', __name__,
                                 template_folder='templates',
                                 static_folder='static',
                                 url_prefix='/api/v1/device-util')


@device_util_bp.route('/test-push-notification', methods=['post'])
@device_util_bp.route('/test-terminal-push-notification', methods=['post'])
@request_headers
@request_values
def test_push_notification(request_headers, request_values):
    title               = request_values.get('title')
    message             = request_values.get('message')
    language_code       = request_values.get('languageCode', DEFAULT_LANGUAGE)
    device_type         = request_values.get('device_type')
    platform            = request_headers.get('x-platform')
    activation_code     = request_values.get('activation_code')
    
    logger.info('platform=%s', platform)
    logger.info('activation_code=%s', activation_code)
    logger.info('title=%s', title)
    logger.info('message=%s', message)
    logger.info('device_type=%s', device_type)
    logger.info('language_code=%s', language_code)
    
    if is_not_empty(activation_code) and is_not_empty(device_type) and is_not_empty(title) and is_not_empty(message) and is_not_empty(platform):
        db_client = create_db_client(caller_info="test_push_notification")
        with db_client.context():
            if device_type=='loyalty':
                device_setting = LoyaltyDeviceSetting.get_by_activation_code(activation_code)
        
        if device_setting:
            logger.debug('Found device setting')
            logger.info('device_setting.device_details=%s', device_setting.device_details)
            device_details_by_platform = device_setting.device_details.get(platform)
            if is_not_empty(device_details_by_platform):
                device_token = device_details_by_platform.get('device_token')
                logger.info('device_token=%s', device_token)
                if is_not_empty(device_token):
                    create_push_notification(
                                title_data      = title, 
                                message_data    = message,
                                speech          = message,
                                device_token    = device_token,
                                language_code   = language_code,
                                
                            )
                else:
                    logger.debug('Device token is empty')
            else:
                logger.debug('Device details is empty')
        
    return create_rest_message(status_code=StatusCode.OK) 

@device_util_bp.route('/test-device-push-notification', methods=['post'])
@request_values
def test_device_push_notification(request_values):
    
    title               = request_values.get('title')
    message             = request_values.get('message')
    language_code       = request_values.get('languageCode', DEFAULT_LANGUAGE)
    device_token        = request_values.get('device_token')
    
    logger.info('device_token=%s', device_token)
    logger.info('title=%s', title)
    logger.info('message=%s', message)
    logger.info('language_code=%s', language_code)
    
    if is_not_empty(device_token) and is_not_empty(language_code) and is_not_empty(title) and is_not_empty(message):
        create_push_notification(
                                title_data      = title, 
                                message_data    = message,
                                speech          = message,
                                device_token    = device_token,
                                language_code   = language_code,
                                
                            )
        
    return create_rest_message(status_code=StatusCode.OK) 
    

        
        
        
        


from wtforms import StringField, PasswordField, validators, DateField
from trexlib.forms.base_forms import ValidationBaseForm
from trexlib.libs.wtforms import validators as custom_validator
from flask_babel import gettext
from wtforms.fields.core import FloatField


class UserMinForm(ValidationBaseForm):
    name                = StringField(gettext('Name'), validators=[
                                        validators.InputRequired(gettext('Name is required')),
                                        validators.Length(min=3, max=300, message='Name length must be within 3 and 300 characters'),
                                        
                                        ]
                                        )
    email               = StringField('Email', validators=[
                                        #validators.Email(gettext("Please enter valid email address.")),
                                        #validators.InputRequired(gettext('Email is required')),
                                        custom_validator.RequiredIfOtherFieldsValueIsEmpty(
                                                        ['mobile_phone'],
                                                        message=gettext("Either email or mobile phone is required")),
                                        
                                        ]
                                        )
    mobile_phone        = StringField('Mobile Phone', validators=[
                                        custom_validator.RequiredIfOtherFieldsValueIsEmpty(
                                                        ['email'],
                                                        message=gettext("Either email or mobile phone is required")),
                                        ]
                                        )
    
    gender              = StringField('Gender', validators=[
                                        #validators.InputRequired(gettext('Gender is required')),
                                        ]
                                        )
    referrer_code       = StringField('Referrer Code', validators=[
                                        ]
                                        )
    
    birth_date          = StringField('Date of Birth', validators=[
                                        #validators.InputRequired(gettext('Birth date is required')),
                                        ]
                                        )
    
    
    
class UserRegistrationForm(UserMinForm):
    password            = StringField(gettext('Password'), validators=[
                                        validators.InputRequired(gettext('Password is required')),
                                        
                                        ]
                                        ) 
    status              = StringField('Status', validators=[
                                        
                                        ]
                                        ) 
    
class UserUpdateForm(UserMinForm):
    reference_code      = StringField('User Reference Code', validators=[
                                        validators.InputRequired(gettext('User Reference Code is required')),
                                        ]
                                        )
    status      = StringField('User Status', validators=[
                                        ]
                                        )
    
class UserStatusForm(ValidationBaseForm):
    status      = StringField('User Status', validators=[
                                        validators.InputRequired(gettext('User status is required')),
                                        ]
                                        )          
    
class OutletReviewsForm(ValidationBaseForm):
    outlet_key               = StringField('Outlet Key', validators=[
                                        validators.InputRequired(gettext('Outlet key is required')),
                                        ]
                                        )
    
    food_score               = FloatField('Food Score', validators=[
                                        validators.InputRequired(gettext('Food score is required')),
                                        ]
                                        )
    
    service_score               = FloatField('Service Score', validators=[
                                        validators.InputRequired(gettext('Service score is required')),
                                        ]
                                        )
    
    ambience_score          = FloatField('Ambience Score', validators=[
                                        validators.InputRequired(gettext('Ambience score is required')),
                                        ]
                                        )
    
    value_for_money_score   = FloatField('Value for money Score', validators=[
                                        validators.InputRequired(gettext('Value for money score is required')),
                                        ]
                                        )    


from flask import Blueprint
import logging
from trexapi.decorators.api_decorators import auth_token_required
from trexlib.utils.string_util import is_not_empty, is_empty
from trexmodel.models.datastore.pos_models import POSSetting
from trexmodel.utils.model.model_util import create_db_client
from datetime import datetime
from trexlib.libs.flask_wtf.request_wrapper import request_headers


#logger = logging.getLogger('api')
logger = logging.getLogger('debug')

transaction_api_bp = Blueprint('transaction_api_bp', __name__,
                                 template_folder='templates',
                                 static_folder='static',
                                 url_prefix='/api/v1/transaction')

@transaction_api_bp.route('/ping', methods=['GET'])
def ping():
    return 'pong',200

@transaction_api_bp.route('/reference-code/<reference_code>/read', methods=['GET'])
@auth_token_required
@request_headers
def list_customer_transaction():
    
    db_client = create_db_client(caller_info="list_customer_transaction")
    
    


'''
Created on 8 Jul 2021

@author: jacklok
'''
from flask import request
from trexlib.utils.crypto_util import decrypt_json
from datetime import datetime, timedelta
from trexconf import conf as api_conf
from trexlib.utils.crypto_util import encrypt_json
import six
from six import string_types
from trexmodel.models.datastore.merchant_models import MerchantAcct, Outlet
from trexlib.libs.google.gae_app_logger import AppLogger

logger = AppLogger.get_logger('helper')


def get_logged_in_api_username():
    auth_token  = request.headers.get('x-auth-token')
    username    = None
    try:
        auth_details_json = decrypt_json(auth_token)
    except:
        logger.error('Failed to decrypt authenticated token')
        
    logger.debug('auth_details_json=%s', auth_details_json)
    
    if auth_details_json:
        username = auth_details_json.get('username')
        
    return username

def get_logged_in_outlet_key(): 
    outlet_key      = request.headers.get('x-outlet-key')
    return outlet_key

def get_logged_in_outlet(): 
    outlet_key      = request.headers.get('x-outlet-key')
    return Outlet.get_or_read_from_cache(outlet_key)

def get_logged_in_merchant_acct(): 
    acct_id         = request.headers.get('x-acct-id')
    merchant_acct   = MerchantAcct.get_or_read_from_cache(acct_id)
    
    return merchant_acct

def generate_user_auth_token(acct_id, reference_code, device_id):
    expiry_datetime = datetime.now() + timedelta(minutes = int(api_conf.API_TOKEN_EXPIRY_LENGTH_IN_MINUTE))
    
    logger.debug('expiry_datetime=%s', expiry_datetime)
    
    token_content =  {
                        'acct_id'           : acct_id,
                        'reference_code'    : reference_code,
                        'expiry_datetime'   : expiry_datetime.strftime('%d-%m-%Y %H:%M:%S'),
                        'device_id'         : device_id,
                        }
    
    logger.debug('token_content=%s', token_content)
    
    return token_content

def encrypt_user_auth_token(auth_token):
    logger.debug('auth_token=%s', auth_token)
    
    return encrypt_json(auth_token)

class StatusCode(object):
    OK                          = 200
    CREATED                     = 201
    ACCEPTED                    = 202
    NO_CONTENT                  = 204
    RESET_CONTENT               = 205
    BAD_REQUEST                 = 400
    UNAUTHORIZED                = 401
    FORBIDDEN                   = 403
    NOT_FOUND                   = 404
    METHOD_NOT_ALLOW            = 405
    PRECONDITION_FAILED         = 412
    RESOURCE_LOCKED             = 423
    INTERNAL_SERVER_ERROR       = 500
    SERVICE_NOT_AVAILABLE       = 503
    GATEWAY_TIMEOUT             = 504
    HTTP_VERSION_NOT_SUPPORT    = 505
    # add more status code according to your need

def create_api_message(message=None, status_code=StatusCode.BAD_REQUEST, **kwargs):
    
    request_content_type = request.headers.get('content-type')
    reply_message = {}
    logger.debug('create_api_message: request_content_type=%s', request_content_type)
    logger.debug('create_api_message: message=%s', message)
    
    if kwargs is not None:
        for key, value in six.iteritems(kwargs):
            reply_message[key] = value
        
    if message:
        if isinstance(message, string_types):
            reply_message['msg'] = [message]
        
        elif isinstance(message, (tuple, list)):
            reply_message['msg'] = message
            
        elif isinstance(message, dict):
            reply_message['msg'] = message['msg']    
    #else:
    #    reply_message['msg'] = []
    
    logger.debug('reply_message=%s', reply_message)
    
    return reply_message, status_code


'''
Created on 29 Jul 2025

@author: jacklok
'''

from flask import Blueprint
import logging
from trexlib.utils.string_util import is_not_empty
from trexapi.utils.api_helpers import StatusCode, create_api_message
from trexmodel.utils.model.model_util import create_db_client
from trexlib.libs.flask_wtf.request_wrapper import request_values, outlet_key,\
    request_args
from trexapi.decorators.api_decorators import auth_token_required
from trexmodel.models.datastore.merchant_models import Outlet
from trexmodel.models.datastore.transaction_models import SalesTransaction
from trexconf import conf
from datetime import datetime
from flask.json import jsonify

#logger = logging.getLogger('api')
logger = logging.getLogger('target_debug')

outlet_report_api_bp = Blueprint('outlet_report_api_bp', __name__,
                                 template_folder='templates',
                                 static_folder='static',
                                 url_prefix='/api/v1/outlet-report')

@outlet_report_api_bp.route('/sales-report-by-date', methods=['GET'])
@auth_token_required
@outlet_key
@request_args
def get_outlet_today_sales_report(outlet_key, request_args):
    db_client = create_db_client(caller_info="get_outlet_today_sales_report")
    transactions_list = []
    
    enquiry_date_str            = request_args.get('enquiry_date')
    
    logger.debug('enquiry_date=%s', enquiry_date_str)
    logger.debug('outlet_key=%s', outlet_key)
    
    with db_client.context():
        outlet = Outlet.fetch(outlet_key)
        
        
    
    if outlet:
        dict_properties  = ['transaction_id', 'invoice_id', 'remarks', 'tax_amount', 'transact_amount', 
                           'transact_datetime', 
                           ]
        with db_client.context():
            enquiry_date                = datetime.strptime(enquiry_date_str, '%d-%m-%Y')
                
            logger.debug('enquiry_date=%s', enquiry_date)
            result       = SalesTransaction.list_transaction_by_date(enquiry_date, transact_outlet=outlet, including_reverted_transaction=False, limit=conf.MAX_FETCH_RECORD)
            
            for r in result:
                transactions_list.append(r.to_dict(dict_properties=dict_properties,  date_format="%d-%m-%Y", datetime_format="%d-%m-%Y %H:%M:%S"))
        
        logger.debug('transactions_list=%s', transactions_list)
        
        return create_api_message(result=transactions_list,  status_code=StatusCode.OK)
        
    else:
        return create_api_message(status_code=StatusCode.BAD_REQUEST)   



'''
Created on 28 Jul 2025

@author: jacklok
'''

from flask import Blueprint, request, url_for
import logging
from trexlib.utils.string_util import is_not_empty
from trexapi.utils.api_helpers import StatusCode, create_api_message
from trexmodel.utils.model.model_util import create_db_client
from trexlib.libs.flask_wtf.request_wrapper import request_values
from trexlib.utils.log_util import get_tracelog
from trexmodel.models.datastore.support_models import ErrorReport


logger = logging.getLogger('api')
#logger = logging.getLogger('debug')

support_api_bp = Blueprint('support_api_bp', __name__,
                                 template_folder='templates',
                                 static_folder='static',
                                 url_prefix='/api/v1/support')

@support_api_bp.route('/report-error', methods=['POST'])
@request_values
def report_error(request_values):
    platform        = request_values.get('platform')
    activation_code = request_values.get('activation_code')
    error_message   = request_values.get('error_message')
    stack_trace     = request_values.get('stack_trace')
    
    logger.info('request_values=%s', request_values)
    
    logger.info('activation_code=%s', activation_code)
    logger.info('platform=%s', platform)
    logger.info('error_message=%s', error_message)
    logger.info('stack_trace=%s', stack_trace)
    
    if is_not_empty(platform) and is_not_empty(error_message):
        db_client                           = create_db_client(caller_info="update_device_details")
        
        with db_client.context():
            ErrorReport.create(platform, error_message, activation_code=activation_code, stack_trace=stack_trace)
            
        return create_api_message(status_code=StatusCode.ACCEPTED)
    else:
        return create_api_message('Missing required data', status_code=StatusCode.BAD_REQUEST)       



from flask import Blueprint
import logging
from trexmodel.utils.model.model_util import create_db_client
from trexmodel.models.datastore.merchant_models import MerchantAcct
from trexmodel.models.datastore.customer_models import Customer
from trexmodel.models.datastore.reward_models import CustomerEntitledTierRewardSummary, CustomerPointReward
from trexapi.decorators.api_decorators import auth_token_required
from flask.json import jsonify
from trexapi.utils.api_helpers import StatusCode, create_api_message
from trexlib.libs.flask_wtf.request_wrapper import request_headers

customer_reward_api_bp = Blueprint('customer_reward_api_bp', __name__,
                                 template_folder='templates',
                                 static_folder='static',
                                 url_prefix='/api/v1/customers/reward')

logger = logging.getLogger('target_debug')


@customer_reward_api_bp.route('/ping', methods=['GET'])
def ping():
    return create_api_message('OK', status_code=StatusCode.OK)

@customer_reward_api_bp.route('/reference-code/<reference_code>', methods=['GET'])
@auth_token_required
@request_headers
def read_customer_reward_summary(request_headers, reference_code):
    acct_id  = request_headers.get('x-acct-id')
    
    logger.debug('reference_code=%s', reference_code)
    logger.debug('acct_id=%s', acct_id)
    
    tier_rewards    = []
    
    db_client = create_db_client(caller_info="read_customer_reward_summary")
    with db_client.context():
        merchant_acct = MerchantAcct.fetch(acct_id)
        customer = Customer.get_by_reference_code(reference_code, merchant_acct)
        if customer:
            #customer_vouchers_list          = CustomerEntitledVoucher.list_all_by_customer(customer)
            customer_tier_reward_summary    = CustomerEntitledTierRewardSummary.list_tier_reward_summary_by_customer(customer)
            CustomerPointReward.list_by_customer(customer)
            '''
            if customer_vouchers_list:
                for v in customer_vouchers_list:
                    vouchers_list.append(v.to_dict())
            '''        
            if customer_tier_reward_summary:
                for v in customer_tier_reward_summary:
                    tier_rewards.append(v.to_dict())
    
    
    
    result = {
            'reference_code'    : reference_code,
            #'tier_rewards'      : tier_rewards,
            #'reward_summary'    : customer.reward_summary,
            #'prepaid_summary'   : customer.prepaid_summary,
            #'voucher_summary'   : customer.entitled_voucher_summary,
            }
    
    if tier_rewards:
        result['tier_rewards'] = tier_rewards
    
    if customer.reward_summary:
        result['reward_summary'] = customer.reward_summary
        
    if customer.prepaid_summary:
        result['prepaid_summary'] = customer.prepaid_summary
    
    if customer.entitled_voucher_summary:
        result['voucher_summary'] = customer.entitled_voucher_summary            
    
    return jsonify(result)

@customer_reward_api_bp.route('/reference-code/<reference_code>/give', methods=['POST'])
@auth_token_required
@request_headers
def give_customer_reward(request_headers, reference_code):
    acct_id  = request_headers.get('x-acct-id')
    logger.debug('reference_code=%s', reference_code)
    logger.debug('acct_id=%s', acct_id)
    #vouchers_list   = []
    tier_rewards    = []
    
    db_client = create_db_client(caller_info="give_customer_reward")
    with db_client.context():
        merchant_acct = MerchantAcct.fetch(acct_id)
        customer = Customer.get_by_reference_code(reference_code, merchant_acct)
        
        if customer:
            #customer_vouchers_list          = CustomerEntitledVoucher.list_all_by_customer(customer)
            customer_tier_reward_summary    = CustomerEntitledTierRewardSummary.list_tier_reward_summary_by_customer(customer)
            CustomerPointReward.list_by_customer(customer)
            
            if customer_tier_reward_summary:
                for v in customer_tier_reward_summary:
                    tier_rewards.append(v.to_dict())
    
    
    
    result = {
            'reference_code'    : reference_code,
            #'vouchers'          : vouchers_list,
            'tier_rewards'      : tier_rewards,
            'reward_summary'    : customer.reward_summary,
            'prepaid_summary'   : customer.prepaid_summary,
            'voucher_summary'   : customer.entitled_voucher_summary,
            }
    
    return jsonify(result)

    


from flask import Blueprint, request
import logging
from trexmodel.utils.model.model_util import create_db_client
from trexmodel.models.datastore.merchant_models import MerchantAcct, Outlet,\
    MerchantUser
from trexmodel.models.datastore.customer_models import Customer,\
    CustomerMembership, CustomerTierMembership,\
    CustomerTierMembershipAccumulatedRewardSummary
from trexapi.decorators.api_decorators import auth_token_required
#from trexadmin.libs.decorators import elapsed_time_trace
from trexmodel.models.datastore.model_decorators import model_transactional
from trexmodel.models.datastore.membership_models import MerchantMembership
from trexapi.utils.api_helpers import get_logged_in_api_username, get_logged_in_merchant_acct, get_logged_in_outlet,\
    create_api_message, StatusCode
from trexmodel.models.datastore.transaction_models import CustomerTransaction
from trexmodel.models.datastore.helper.reward_transaction_helper import check_giveaway_reward_for_membership_purchase_transaction
from datetime import datetime
from trexlib.libs.flask_wtf.request_wrapper import request_headers,\
    request_values
from trexlib.utils.string_util import is_not_empty
from flask.json import jsonify


customer_membership_api_bp = Blueprint('customer_membership_api_bp', __name__,
                                 template_folder='templates',
                                 static_folder='static',
                                 url_prefix='/api/v1/customer-membership')

logger = logging.getLogger('debug')


@customer_membership_api_bp.route('/ping', methods=['GET'])
def ping():
    return create_api_message('OK', status_code=StatusCode.OK)

@customer_membership_api_bp.route('/customer/reference-code/<reference_code>', methods=['GET'])
@auth_token_required
@request_headers
def read_customer_membership(request_headers, reference_code):
    acct_id         = request_headers.get('x-acct-id')
    
    
    logger.debug('reference_code=%s', reference_code)
    logger.debug('acct_id=%s', acct_id)
    
    db_client = create_db_client(caller_info="with db_client.context():")
    
    membership_details_list = []
    
    with db_client.context():
        merchant_acct = MerchantAcct.fetch(acct_id)
        customer = Customer.get_by_reference_code(reference_code, merchant_acct)
        if customer:
            customer_memberships_list = CustomerMembership.list_active_by_customer(customer)
            
            if customer_memberships_list:
                for custome_membership in customer_memberships_list:
                    merchant_membership_program = custome_membership.merchant_membership_entity
                    membership_details = {
                                        'label'                     : merchant_membership_program.label,
                                        'desc'                      : merchant_membership_program.desc,
                                        'terms_and_conditions'      : merchant_membership_program.terms_and_conditions,
                                        'entitled_date'             : datetime.strftime(custome_membership.entitled_date, '%d-%m-%Y'),
                                        'expiry_date'               : datetime.strftime(custome_membership.expiry_date, '%d-%m-%Y'),
                                        
                            }
                    membership_details_list.append(membership_details)
                
            customer_tier_memberships_list = CustomerTierMembership.list_active_by_customer(customer)
            
            if customer_tier_memberships_list:
                for custome_membership in customer_tier_memberships_list:
                    merchant_membership_program = custome_membership.merchant_tier_membership_entity
                    membership_details = {
                                        'label'                     : merchant_membership_program.label,
                                        'desc'                      : merchant_membership_program.desc,
                                        'terms_and_conditions'      : merchant_membership_program.terms_and_conditions,
                                        'entitled_date'             : datetime.strftime(custome_membership.entitled_date, '%d-%m-%Y'),
                                        'expiry_date'               : datetime.strftime(custome_membership.expiry_date, '%d-%m-%Y'),
                                        
                            }
                    membership_details_list.append(membership_details)
    
    return jsonify(membership_details_list)            
        

@customer_membership_api_bp.route('/customer/reference-code/<reference_code>/assign-membership', methods=['POST'])
@auth_token_required
@request_values
@request_headers
#@elapsed_time_trace(trace_key="assign_membership")
def assign_membership(request_values, request_headers, reference_code):
    acct_id         = request_headers.get('x-acct-id')
    outlet_key      = request_headers.get('x-outlet-key')
    
    logger.debug('reference_code=%s', reference_code)
    logger.debug('outlet_key=%s', outlet_key)
    logger.debug('acct_id=%s', acct_id)
    
    merchant_membership_key     = request_values.get('merchant_membership_key')
    number_of_year              = request_values.get('number_of_year')
    
    if is_not_empty(number_of_year):
        number_of_year = int(number_of_year)
    
    if is_not_empty(merchant_membership_key):
    
        db_client = create_db_client(caller_info="assign_membership")
        customer_membership = None
        
        with db_client.context():
            merchant_acct = MerchantAcct.fetch(acct_id)
            customer = Customer.get_by_reference_code(reference_code, merchant_acct)
            if customer:
                merchant_membership = MerchantMembership.fetch(merchant_membership_key)
                            
                customer_membership = CustomerMembership.get_by_customer_and_merchant_membership(customer, merchant_membership)
                
        if customer_membership:
            return create_api_message('The customer membership has already been assigned.', status_code=StatusCode.BAD_REQUEST)
        else:
            with db_client.context():    
                merchant_username       = get_logged_in_api_username()
                assigned_by             = MerchantUser.get_by_username(merchant_username)
                assigned_outlet         = Outlet.fetch(outlet_key)
                if customer:
                    __assign_membership(customer, merchant_membership, assigned_by, assigned_outlet, number_of_year=number_of_year)
                
                
        
            return create_api_message(status_code=StatusCode.OK)
            
        
        return create_api_message(status_code=StatusCode.OK)
    else:
        return create_api_message('Missing membership program key', status_code=StatusCode.BAD_REQUEST)

@model_transactional(desc="assign customer membership")
def __assign_membership(customer, merchant_membership, assigned_by, assigned_outlet, number_of_year=None):
    customer_membership = CustomerMembership.create(customer, merchant_membership, assigned_by=assigned_by, assigned_outlet=assigned_outlet, number_of_year=number_of_year)
    
    customer_transaction = CustomerTransaction.create_membership_purchase_transaction(
                                customer, customer_membership, 
                                system_remarks= "Joined Membership", 
                                transact_outlet=assigned_outlet, 
                                transact_by=assigned_by, 
                                )
    
    check_giveaway_reward_for_membership_purchase_transaction(customer, customer_transaction)
    
@customer_membership_api_bp.route('/customer/reference-code/<reference_code>/renew-membership', methods=['POST'])
@auth_token_required
#@elapsed_time_trace(trace_key="renew_membership")
def renew_membership(reference_code):
    
    merchant_acct = get_logged_in_merchant_acct() 
    outlet_key      = request.headers.get('x-outlet-key')
    logger.debug('reference_code=%s', reference_code)
    logger.debug('outlet_key=%s', outlet_key)
    
    merchant_membership_key = request.args.get('merchant_membership_key') or request.form.get('merchant_membership_key') or request.json.get('merchant_membership_key')
    
    db_client = create_db_client(caller_info="renew_membership")
    customer_membership = None
    
    with db_client.context():
        merchant_acct   = get_logged_in_merchant_acct()
        renewed_outlet  = get_logged_in_outlet()
        customer = Customer.get_by_reference_code(reference_code, merchant_acct)
        if customer:
            merchant_membership     = MerchantMembership.fetch(merchant_membership_key)
            merchant_username       = get_logged_in_api_username()
            renewed_by              = MerchantUser.get_by_username(merchant_username)
            #CustomerMembership.renew(customer, merchant_membership, renewed_datetime=datetime.utcnow())
            try:
                customer_membership = __renew_membership(customer, merchant_membership, renewed_by, renewed_outlet)
            except Exception as e:
                return create_api_message(e.args[0], status_code=StatusCode.BAD_REQUEST)
            
    if customer_membership:
        return create_api_message('Customer membership have been renew successfully', status_code=StatusCode.OK)
    else:
        return create_api_message('Customer membership is not found', status_code=StatusCode.BAD_REQUEST)
        
    
    return create_api_message(status_code=StatusCode.OK)    


@model_transactional(desc="renew customer membership")
def __renew_membership(customer, merchant_membership, renewed_by, renewed_outlet):
    customer_membership = CustomerMembership.renew(customer, merchant_membership, renewed_datetime=datetime.utcnow(), renewed_outlet=renewed_outlet, renewed_by=renewed_by)
    
    customer_transaction = CustomerTransaction.create_membership_purchase_transaction(
                                customer, customer_membership, 
                                system_remarks= "Renewed Membership", 
                                transact_outlet=renewed_outlet, 
                                transact_by=renewed_by, 
                                )
    
    check_giveaway_reward_for_membership_purchase_transaction(customer, customer_transaction)
    
    return  customer_membership   

@customer_membership_api_bp.route('/customer/reference-code/<reference_code>/remove-tier-membership', methods=['DELETE'])
@auth_token_required
#@elapsed_time_trace(trace_key="renew_membership")
def remove_customer_tier_membership(reference_code):
    db_client = create_db_client(caller_info="renew_membership")
    customer_membership = None
    is_deleted = False
    
    with db_client.context():
        merchant_acct   = get_logged_in_merchant_acct()
        customer        = Customer.get_by_reference_code(reference_code, merchant_acct)
        if customer:
            is_deleted = CustomerTierMembership.remove_by_customer(customer)
            CustomerTierMembershipAccumulatedRewardSummary.delete_all_for_customer(customer)
            customer.tier_membership = None
            customer.put()
            
    if is_deleted:
        return create_api_message('Customer tier membership have been deleted', status_code=StatusCode.OK)
    else:
        return create_api_message('Customer tier membership not found', status_code=StatusCode.OK)
        
    


'''
Created on 29 Nov 2023

@author: jacklok
'''
from flask import Blueprint
import logging
from trexmodel.utils.model.model_util import create_db_client
from datetime import datetime
from trexlib.utils.string_util import is_not_empty
from trexmodel.models.datastore.user_models import User
from trexmodel.models.datastore.merchant_models import MerchantAcct, Outlet
from trexapi.utils.api_helpers import StatusCode,\
    create_api_message
from trexapi.decorators.api_decorators import user_auth_token_required,\
    user_auth_token_required_pass_reference_code
from flask.json import jsonify
from trexmodel.models.datastore.rating_models import OutletRating,\
    OutletRatingResult, MerchantRatingResult, TransactionRating
from trexlib.libs.flask_wtf.request_wrapper import request_values
from trexmodel.models.datastore.transaction_models import CustomerTransaction

rating_api_bp = Blueprint('rating_api_bp', __name__,
                                 template_folder='templates',
                                 static_folder='static',
                                 url_prefix='/api/v1/rating')

#logger = logging.getLogger('api')
logger = logging.getLogger('target_debug')
'''
@rating_api_bp.route('/outlet/<outlet_key>/review', methods=['POST'])
@user_auth_token_required_pass_reference_code
@request_values
def give_outlet_rating(request_values, reference_code, outlet_key):
    logger.debug('---give_outlet_rating---')
    
    service_rating              = request_values.get('service_rating')
    ambience_rating             = request_values.get('ambience_rating')
    food_rating                 = request_values.get('food_rating')
    value_rating                = request_values.get('value_rating')
    
    
    db_client = create_db_client(caller_info="give_outlet_rating")
    
    logger.debug('give_outlet_rating: user account by reference code=%s', reference_code)
    logger.debug('give_outlet_rating: outlet_key=%s', outlet_key)
    
    with db_client.context():
        user_acct       = User.get_by_reference_code(reference_code)
        outlet          = Outlet.fetch(outlet_key)
        merchant_acct   = outlet.merchant_acct_entity
    
    if user_acct and outlet:
        with db_client.context():
            OutletRating.create(user_acct, outlet, 
                                    merchant_acct   = merchant_acct,
                                    service_rating  = service_rating, 
                                    ambience_rating = ambience_rating, 
                                    food_rating     = food_rating, 
                                    value_rating    = value_rating,
                                    )
            MerchantRatingResult.update(merchant_acct)
        
        return create_api_message(status_code=StatusCode.OK)
    else:
        return create_api_message(status_code=StatusCode.BAD_REQUEST)
    
'''  

@rating_api_bp.route('/outlet/<outlet_key>/review', methods=['GET'])
def read_outlet_rating(outlet_key):   
    logger.debug('give_outlet_rating: outlet_key=%s', outlet_key)
    
    db_client = create_db_client(caller_info="read_outlet_rating")
    
    with db_client.context():
        outlet                  = Outlet.fetch(outlet_key) 
        merchant_acct           = outlet.merchant_acct_entity
        outlet_rating_result    = OutletRatingResult.get_by_outlet(outlet)
        industry                = merchant_acct.industry
    
    
    if outlet_rating_result:
        result = {
                'industry'              : industry,
                'rating_result'         :{
                                        'score'                 : outlet_rating_result.score,
                                        'total_rating_count'    : outlet_rating_result.total_rating_count,
                                        'reviews_details'       : outlet_rating_result.rating_result,
                                        }
            }
    else:
        result = {
                'industry'              : industry,
                'rating_result'         : {
                                            'score'                 : outlet_rating_result.score,
                                            'total_rating_count'    : outlet_rating_result.total_rating_count,
                                            'reviews_details'       : outlet_rating_result.rating_result,
                    }
                
            }
    logger.debug('result=%s', result)    
    #return create_api_message(status_code=StatusCode.OK, result = result)
    return jsonify(result)
    
@rating_api_bp.route('/merchant/<merchant_acct_key>/review', methods=['GET'])
def read_merchant_rating(merchant_acct_key):   
    logger.debug('read_merchant_rating: merchant_acct_key=%s', merchant_acct_key)
    
    db_client = create_db_client(caller_info="read_merchant_rating")
    
    with db_client.context():
        merchant_acct           = MerchantAcct.fetch(merchant_acct_key)
        industry                = merchant_acct.industry 
        merchant_rating_result  = MerchantRatingResult.get_by_merchant_acct(merchant_acct)
    
    
    if merchant_rating_result:
        result = {
                'industry'              : industry,
                'rating_result'         :{
                                        'score'                 : merchant_rating_result.score,
                                        'total_rating_count'    : merchant_rating_result.total_rating_count,
                                        'reviews_details'       : merchant_rating_result.rating_result,
                                        }
            }
    else:
        result = {
                'industry'              : industry,
                'rating_result'         : {
                                            'score'                 : 0,
                                            'total_rating_count'    : 0,
                                            'reviews_details'       : 0,
                    }
                
            }
        
    logger.debug('result=%s', result)    
    
    #return create_api_message(status_code=StatusCode.OK, result = result)
    return jsonify(result)
        
    
@rating_api_bp.route('/outlet/<outlet_key>/update', methods=['POST'])
@user_auth_token_required
@request_values
def update_outlet_rating(request_values, outlet_key):   
    logger.debug('update_outlet_rating: outlet_key=%s', outlet_key)
    updated_datetime_from       = request_values.get('updated_datetime_from')
    
    db_client = create_db_client(caller_info="update_outlet_rating")
    
    if is_not_empty(updated_datetime_from):
        updated_datetime_from = datetime.strptime(updated_datetime_from, '%d-%m-%Y %H:%M')
    
    with db_client.context():
        outlet      = Outlet.fetch(outlet_key) 
        OutletRatingResult.update(outlet, updated_datetime_from)
    
    
    return create_api_message(triggered_datetime=datetime.now(), status_code=StatusCode.OK)

@rating_api_bp.route('/transaction-id/<transaction_id>/review', methods=['POST'])
@user_auth_token_required_pass_reference_code
@request_values
def give_transaction_rating(request_values, reference_code, transaction_id):
    logger.debug('---give_transaction_rating---')
    
    rating_result              = request_values.get('rating_result')
    remarks                    = rating_result.get('remarks','')
    
    if 'remarks' in rating_result:
        del rating_result['remarks']
    
    db_client = create_db_client(caller_info="give_outlet_rating")
    
    logger.debug('give_transaction_rating: user account by reference code=%s', reference_code)
    logger.debug('give_transaction_rating: transaction_id=%s', transaction_id)
    logger.debug('give_transaction_rating: remarks=%s', remarks)
    logger.debug('give_transaction_rating: rating_result=%s', rating_result)
    
    with db_client.context():
        user_acct       = User.get_by_reference_code(reference_code)
        
    
    if user_acct and transaction_id:
        
        with db_client.context():
            TransactionRating.create(user_acct, transaction_id, 
                                rating_result   = rating_result, 
                                remarks         = remarks,
                                for_testing     = False,
                                rating_datetime = datetime.utcnow(),
                                )
            
        
        return create_api_message(status_code=StatusCode.OK)
    else:
        return create_api_message(status_code=StatusCode.BAD_REQUEST)  
    
@rating_api_bp.route('/transaction-id/<transaction_id>/review', methods=['GET'])
@user_auth_token_required_pass_reference_code
def read_transaction_rating(reference_code, transaction_id):
    logger.debug('---read_transaction_rating---')
    
    db_client = create_db_client(caller_info="read_transaction_rating")
    
    logger.debug('read_transaction_rating: user account by reference code=%s', reference_code)
    logger.debug('read_transaction_rating: transaction_id=%s', transaction_id)
    
    
    transaction_review_json = {}
    if transaction_id:
        with db_client.context():
            transaction_review = TransactionRating.get_by_transaction_id(transaction_id)
            logger.debug('read_transaction_rating: transaction_review=%s', transaction_review)
            if transaction_review:
                transaction_review_json = transaction_review.to_dict()
                transaction_review_json['rating_result']['remarks'] = transaction_review_json['remarks']
            else:
                customer_transaction = CustomerTransaction.get_by_transaction_id(transaction_id)
                if customer_transaction:
                    transaction_review_json['industry'] = customer_transaction.industry
                    
                    
        #return create_api_message(status_code=StatusCode.OK, transaction_review=transaction_review_json)
        logger.debug('transaction_review_json=%s', transaction_review_json)
        return jsonify(transaction_review_json)
    else:
        return create_api_message(status_code=StatusCode.BAD_REQUEST)      


@rating_api_bp.route('/merchant/<merchant_acct_key>/update', methods=['POST'])
@request_values
def update_merchant_rating(request_values, merchant_acct_key):   
    logger.debug('update_merchant_rating: merchant_acct_key=%s', merchant_acct_key)
    updated_datetime_from       = request_values.get('updated_datetime_from')
    
    db_client = create_db_client(caller_info="update_merchant_rating")
    
    if is_not_empty(updated_datetime_from):
        updated_datetime_from = datetime.strptime(updated_datetime_from, '%d-%m-%Y %H:%M')
    
    with db_client.context():
        merchant_acct      = MerchantAcct.fetch(merchant_acct_key) 
        MerchantRatingResult.update(merchant_acct, updated_datetime_from)
    
    
    return create_api_message(triggered_datetime=datetime.now(), status_code=StatusCode.OK)    
    



'''
Created on 7 Jul 2021

@author: jacklok
'''

from flask import Blueprint, request 
import logging
from trexlib.utils.log_util import get_tracelog
from flask_restful import Api
from trexmodel.utils.model.model_util import create_db_client
from flask.json import jsonify
from datetime import datetime
from trexapi.decorators.api_decorators import auth_token_required,\
    outlet_key_required, device_is_activated, show_request_info,\
    user_auth_token_required
from trexlib.utils.string_util import is_not_empty
from trexmodel.models.datastore.customer_models import Customer,\
    CustomerMembership, CustomerTierMembership,\
    CustomerTierMembershipAccumulatedRewardSummary
from trexmodel.models.datastore.user_models import User
from trexapi.utils.api_helpers import StatusCode, create_api_message,\
    get_logged_in_merchant_acct
from trexmodel.models.datastore.merchant_models import Outlet,\
    MerchantAcct, MerchantUser
from trexapi.forms.customer_api_forms import CustomerDetailsNewForm, CustomerDetailsUpdateForm,\
    CustomerSearchForm
from werkzeug.datastructures import ImmutableMultiDict
from trexapi.controllers.user_api_routes import user_details_dict
from trexmodel.models.datastore.transaction_models import CustomerTransaction
from trexapi.utils.api_helpers import get_logged_in_api_username
from trexmodel.models.datastore.redeem_models import CustomerRedemption
from trexmodel.models.datastore.model_decorators import model_transactional
from trexmodel.models.datastore.helper.reward_transaction_helper import revert_redemption,\
    revert_transaction
from trexanalytics.bigquery_upstream_data_config import create_merchant_registered_customer_upstream_for_merchant,\
    create_registered_customer_upstream_for_system
from trexmodel.models.datastore.membership_models import MerchantMembership
from flask_babel import gettext
from trexlib.libs.flask_wtf.request_wrapper import request_json, request_headers,\
    request_values
from trexlib.utils.common.date_util import parse_generic_date
from trexmodel.models.datastore.reward_models import CustomerPointReward,\
    CustomerStampReward, CustomerEntitledVoucher, RevertedCustomerPointReward,\
    RevertedCustomerStampReward, CustomerEntitledTierRewardSummary
from trexmodel.models.datastore.prepaid_models import CustomerPrepaidReward
from trexapi.libs.api_decorator import elapsed_time_trace
from trexmodel.models.merchant_helpers import return_customer_details

logger = logging.getLogger('api')


customer_api_bp = Blueprint('customer_api_base_bp', __name__,
                                 template_folder='templates',
                                 static_folder='static',
                                 url_prefix='/api/v1/customers')

logger = logging.getLogger('target_debug')

customer_api = Api(customer_api_bp)



@customer_api_bp.route('/register', methods=['POST'])
@auth_token_required
@outlet_key_required
@device_is_activated
@request_values
@request_headers
def create_customer(customer_data_in_json, request_headers):
    register_customer_form  = CustomerDetailsNewForm(ImmutableMultiDict(customer_data_in_json))
    
    logger.debug('create_customer: customer_data_in_json=%s', customer_data_in_json)
    
    try:
        if register_customer_form.validate():
            logger.debug('customer registration input is valid')
            db_client = create_db_client(caller_info="create_customer")
            
            is_email_used           = False
            is_mobile_phone_used    = False
            customer_is_exist       = False
            created_customer        = None
            created_user_acct       = None
            
            outlet_key  = request_headers.get('x-outlet-key')
            acct_id     = request_headers.get('x-acct-id')
            
            logger.debug('outlet_key=%s', outlet_key)
            logger.debug('acct_id=%s', acct_id)
            
            with db_client.context():
                merchant_acct           = MerchantAcct.fetch(acct_id)
                outlet                  = Outlet.fetch(outlet_key)
                
                email                   = customer_data_in_json.get('email')
                mobile_phone            = customer_data_in_json.get('mobile_phone')
                merchant_reference_code = customer_data_in_json.get('merchant_reference_code')
                
                birth_date              = customer_data_in_json.get('birth_date')
                            
                if is_not_empty(birth_date):
                    birth_date = parse_generic_date(birth_date)
                
                logger.debug('birth_date=%s', birth_date)

                
                logger.debug('email=%s', email)
                logger.debug('mobile_phone=%s', mobile_phone)
                logger.debug('merchant_reference_code=%s', merchant_reference_code)
                
                if is_not_empty(email):
                    created_user_acct = User.get_by_email(email)
                    
                    #if created_user_acct and created_user_acct.is_email_verified:
                    if created_user_acct:    
                        is_email_used = True
                
                if is_not_empty(mobile_phone):
                    mobile_phone = mobile_phone.replace(" ", "")
                    created_user_acct = User.get_by_mobile_phone(mobile_phone)
                    
                    #if created_user_acct and created_user_acct.is_mobile_phone_verified:
                    if created_user_acct:
                        is_mobile_phone_used = True
                
                
                logger.debug('is_email_used=%s', is_email_used)
                logger.debug('is_mobile_phone_used=%s', is_mobile_phone_used)
                
                if merchant_acct and outlet:
                        
                    logger.debug('merchant_acct.key_in_str=%s', merchant_acct.key_in_str)
                    logger.debug('outlet.merchant_acct_key=%s', outlet.merchant_acct_key)
                    
                    if merchant_acct.key_in_str == outlet.merchant_acct_key: 
                        logger.debug('Valid granted outlet key for merchant acct')
                                                    
                        if is_email_used:
                            created_customer = Customer.get_by_email(email, merchant_acct=merchant_acct)
                        
                        if created_customer is None:    
                            if is_mobile_phone_used:
                                created_customer = Customer.get_by_mobile_phone(mobile_phone, merchant_acct=merchant_acct)
                        
                        if is_email_used or is_mobile_phone_used:
                            if created_customer is None:
                                logger.debug('User account have been created, but customer account is not yet created')
                                
                                created_user_acct = User.update(created_user_acct,
                                                                name                    = customer_data_in_json.get('name'), 
                                                                email                   = customer_data_in_json.get('email'), 
                                                                gender                  = customer_data_in_json.get('gender'),
                                                                birth_date              = birth_date,
                                                                mobile_phone            = mobile_phone, 
                                                                password                = customer_data_in_json.get('password'),
                                                                )
                                
                                created_customer        = Customer.create_from_user(created_user_acct, outlet=outlet, merchant_reference_code=customer_data_in_json.get('merchant_reference_code'))
                            
                            else:
                                customer_is_exist = True
                                logger.warn('Customer account using same email or mobile phone have been created')
                        else:
                            if is_not_empty(merchant_reference_code):
                                checking_customer = Customer.get_by_merchant_reference_code(merchant_reference_code, merchant_acct)
                                if checking_customer is not None:
                                    return create_api_message('Merchant reference code (%s) have been taken' % (merchant_reference_code), status_code=StatusCode.BAD_REQUEST)
                                
                                
                            created_customer        = Customer.create( 
                                                            outlet                  = outlet, 
                                                            name                    = customer_data_in_json.get('name'), 
                                                            email                   = customer_data_in_json.get('email'), 
                                                            gender                  = customer_data_in_json.get('gender'),
                                                            birth_date              = birth_date,
                                                            mobile_phone            = mobile_phone, 
                                                            merchant_reference_code = merchant_reference_code, 
                                                            password                = customer_data_in_json.get('password'),
                                                            )
                            
                            create_merchant_registered_customer_upstream_for_merchant(created_customer)
                            create_registered_customer_upstream_for_system(created_customer)
                            
                        logger.debug('created_customer=%s', created_customer)
                    else:
                        logger.warn('Invalid granted outlet key')
                else:
                    logger.warn('Invalid granted outlet key')
            
                if customer_is_exist:
                    if is_email_used==True:
                        return create_api_message('Email have been taken', status_code=StatusCode.BAD_REQUEST)
                    
                    elif is_mobile_phone_used==True:
                        return create_api_message('Mobile phone have been taken', status_code=StatusCode.BAD_REQUEST)
                    
                
                if created_customer:
                    created_user_acct = created_customer.registered_user_acct
                    response_data = {
                                'customer_key'              : created_customer.key_in_str,
                                'registered_datetime'       : created_customer.registered_datetime.strftime("%d-%m-%Y %H:%M:%S"),
                                'modified_datetime'         : created_customer.modified_datetime.strftime("%d-%m-%Y %H:%M:%S"),
                                'merchant_reference_code'   : created_customer.merchant_reference_code,
                                'reference_code'            : created_customer.reference_code,
                                'referral_code'             : created_customer.referral_code,
                                'registered_outlet_key'     : created_customer.registered_outlet_key,
                                #'merchant_account_key'      : acct_id,
                                #'company_name'              : merchant_acct.company_name,
                                #'outlet_key'                : outlet_key,  
                                
                                }
                
                    logger.debug('response_data=%s', response_data)
                    
                    response_data.update(user_details_dict(created_user_acct))
                    
                    return create_api_message(status_code=StatusCode.OK, **response_data)
                    
                else:
                    return create_api_message(status_code=StatusCode.BAD_REQUEST)
            
        else:
            logger.warn('customer registration input is invalid')
            error_message = register_customer_form.create_rest_return_error_message()
            
            return create_api_message(error_message, status_code=StatusCode.BAD_REQUEST)
    except:
        logger.error('Fail to register customer due to %s', get_tracelog())
        
        return create_api_message(status_code=StatusCode.BAD_REQUEST)

'''
def _return_customer_details(customer):
    customer_details_dict = customer.to_dict(
                                        date_format="%d-%m-%Y", 
                                        datetime_format="%d-%m-%Y %H:%M:%S",
                                        excluded_dict_properties = [
                                                'registered_merchant_acct', 
                                                'memberships_list', 
                                                'tier_membership_key',
                                                'registered_user_acct',
                                                
                                                ],
                                        )
    customer_details_dict['customer_key']               = customer.key_in_str
    
    
    customer_basic_memberships_list = _list_customer_basic_memberships(customer)
    if customer_basic_memberships_list:
        customer_details_dict['basic_memberships'] = customer_basic_memberships_list
    
    customer_tier_membership = _get_tier_membership(customer)
    if customer_tier_membership:
        customer_details_dict['tier_membership']  = customer_tier_membership
    
    if 'entitled_voucher_summary' in customer_details_dict:
        customer_details_dict['voucher_summary']            = customer.entitled_voucher_summary
        del customer_details_dict['entitled_voucher_summary']
    
    return customer_details_dict

def _return_customer_reward(customer):
    customer_reward_dict = customer.to_dict(
                                        date_format="%d-%m-%Y", 
                                        datetime_format="%d-%m-%Y %H:%M:%S",
                                        dict_properties = [
                                                'reward_summary', 'entitled_voucher_summary', 'prepaid_summary', 
                                                'entitled_lucky_draw_ticket_summary', 
                                                
                                                
                                                ],
                                        )
    customer_reward_dict['customer_key']               = customer.key_in_str
    
    
    customer_basic_memberships_list = _list_customer_basic_memberships(customer)
    if customer_basic_memberships_list:
        customer_reward_dict['basic_memberships'] = customer_basic_memberships_list
    
    customer_tier_membership = _get_tier_membership(customer)
    if customer_tier_membership:
        customer_reward_dict['tier_membership']  = customer_tier_membership
    
    return customer_reward_dict

def _return_customer_minimum_details(customer):
    customer_details_dict = customer.to_dict(
                                        date_format="%d-%m-%Y", 
                                        datetime_format="%d-%m-%Y %H:%M:%S",
                                        excluded_dict_properties = [
                                                'registered_merchant_acct', 
                                                'memberships_list', 
                                                'tier_membership_key',
                                                'registered_user_acct',
                                                'kpi_summary',
                                                'entitled_lucky_draw_ticket_summary',
                                                'reward_summary',
                                                'tags_list',
                                                'entitled_membership_reward_summary',
                                                'entitled_birthday_reward_summary',
                                                ],
                                        )
    customer_details_dict['customer_key']               = customer.key_in_str
    
    customer_basic_memberships_list = _list_customer_basic_memberships(customer)
    if customer_basic_memberships_list:
        customer_details_dict['basic_memberships'] = customer_basic_memberships_list
    
    customer_tier_membership = _get_tier_membership(customer)
    if customer_tier_membership:
        customer_details_dict['tier_membership']  = customer_tier_membership
    
    if 'entitled_voucher_summary' in customer_details_dict:
        customer_details_dict['voucher_summary']            = customer.entitled_voucher_summary
        del customer_details_dict['entitled_voucher_summary']
    
    return customer_details_dict
'''
    
@customer_api_bp.route('/delete-customer-tier-membership-and-tier-progression/reference-code/<reference_code>', methods=['DELETE'])
@auth_token_required
def delete_customer_tier_membership_and_accumulated_reward_summary(reference_code):
    
    db_client = create_db_client(caller_info="reset_customer_tier_membership_and_accumulated_reward_summary")
    try:
        with db_client.context():
            merchant_acct   = get_logged_in_merchant_acct()
            if merchant_acct: 
                logger.debug('merchant account company name=%s', merchant_acct.company_name)
            customer = Customer.get_by_reference_code(reference_code, merchant_acct)
            
            if customer:
                logger.debug(f'Found customer={customer.name}')
                
                CustomerTierMembershipAccumulatedRewardSummary.delete_all_for_customer(customer)
                CustomerTierMembership.delete_all_by_customer(customer)
                Customer.remove_tier_membership(customer)
                
                
                
        return create_api_message(status_code=StatusCode.ACCEPTED,
                                               
                                               )
            
            
    except:
        logger.error('Failed due to %s', get_tracelog())
        return create_api_message("Failed to reset customer tier membership and accumulated reward summary", status_code=StatusCode.BAD_REQUEST)
    

@customer_api_bp.route('/tier-membership/reference-code/<reference_code>', methods=['GET'])
@auth_token_required
def read_customer_tier_membership(reference_code):
    
    db_client = create_db_client(caller_info="read_customer_tier_membership")
    tier_membership_details = {}
    try:
        with db_client.context():
            merchant_acct   = get_logged_in_merchant_acct()
            if merchant_acct: 
                logger.debug('merchant account company name=%s', merchant_acct.company_name)
            customer = Customer.get_by_reference_code(reference_code, merchant_acct)
            
            
            if customer :
                logger.debug(f'Found customer={customer.name}')
                
                tier_membership_details['name'] = customer.name
                tier_membership_details['reference_code'] = customer.reference_code
                if customer.tier_membership_entity:
                    customer_tier_membership = CustomerTierMembership.get_by_customer_and_merchant_tier_membership(customer, customer.tier_membership_entity)
                    if customer_tier_membership:
                        merchant_tier_membership = customer.tier_membership_entity
                        tier_membership_details['label'] = merchant_tier_membership.label
                        tier_membership_details['entitled_date'] = datetime.strftime(customer_tier_membership.entitled_date, '%d-%m-%Y')
                        tier_membership_details['expiry_date'] = datetime.strftime(customer_tier_membership.expiry_date, '%d-%m-%Y')
                    
                
                
                
                
        return create_api_message(status_code=StatusCode.OK, tier_membership_details=tier_membership_details
                                               
                                               )
            
            
    except:
        logger.error('Failed due to %s', get_tracelog())
        return create_api_message("Failed to read customer tier membership details", status_code=StatusCode.BAD_REQUEST)    
    
@customer_api_bp.route('/reference-code/<ref_code>', methods=['GET'])
@auth_token_required
@device_is_activated
@request_headers
def read_customer(request_headers, ref_code):
    
    logger.debug('ref_code=%s', ref_code)
    
    if is_not_empty(ref_code):
        db_client = create_db_client(caller_info="read_customer")
        
        acct_id     = request_headers.get('x-acct-id')
        
        with db_client.context():
            merchant_acct = MerchantAcct.fetch(acct_id)
            customer = Customer.get_by_reference_code(ref_code, merchant_acct)
            
        
        if customer:
            with db_client.context():
                customer_details_dict = return_customer_details(customer)
            
            customer_details_json =  jsonify(customer_details_dict)
            
            
            logger.debug('customer_details_json=%s', customer_details_json)
            
            return customer_details_json
            
        else:
            logger.warn('Customer with reference code (%s) is not found', ref_code)
            return create_api_message('Customer with reference code (%s) is not found' % ref_code, 
                                      status_code=StatusCode.BAD_REQUEST)
            
    else:
        return create_api_message(status_code=StatusCode.BAD_REQUEST)

@customer_api_bp.route('/customer-key/<customer_key>/customer-reward', methods=['GET'])
@auth_token_required
@device_is_activated
@elapsed_time_trace(trace_key='read_customer_reward')
def read_customer_reward(customer_key):
    
    logger.debug('customer_key=%s', customer_key)
    
    if is_not_empty(customer_key):
        db_client = create_db_client(caller_info="read_customer_reward")
        
        with db_client.context():
            customer = Customer.fetch(customer_key)
            
        
        if customer:
            with db_client.context():
                customer_details_dict = _return_customer_reward(customer)
            
            customer_details_json =  jsonify(customer_details_dict)
            
            
            logger.debug('customer_details_json=%s', customer_details_json)
            
            return customer_details_json
            
        else:
            logger.warn('Customer with customer key (%s) is not found', customer_key)
            return create_api_message('Customer with customer key (%s) is not found' % customer_key, 
                                      status_code=StatusCode.BAD_REQUEST)
            
    else:
        return create_api_message(status_code=StatusCode.BAD_REQUEST)

@customer_api_bp.route('/reference-code/<reference_code>/read-or-register', methods=['GET'])
@auth_token_required
@device_is_activated
@request_headers
def read_and_register_customer(request_headers, reference_code):
    
    logger.debug('reference_code=%s', reference_code)
    
    if is_not_empty(reference_code):
        db_client = create_db_client(caller_info="read_and_register_customer")
        
        acct_id     = request_headers.get('x-acct-id')
        
        with db_client.context():
            merchant_acct = MerchantAcct.fetch(acct_id)
            customer = Customer.get_by_reference_code(reference_code, merchant_acct)
            
        
        if customer:
            with db_client.context():
                #customer_details_dict = _return_customer_minimum_details(customer)
                customer_details_dict = {
                                        'name': customer.name,
                                        'reference_code': customer.reference_code,
                                        'customer_key'  : customer.key_in_str,
                                        }
            
            customer_details_json =  jsonify(customer_details_dict)
            
            
            logger.debug('customer_details_json=%s', customer_details_json)
            
            return customer_details_json
            
        else:
            logger.warn('Customer with reference code (%s) is not found', reference_code)
            with db_client.context():
                existing_user_acct  = User.get_by_reference_code(reference_code)
                outlet_key  = request_headers.get('x-outlet-key')
                outlet      = Outlet.fetch(outlet_key)
                
                created_customer = __register_user_as_customer(existing_user_acct, outlet)
                
            if created_customer:
                with db_client.context():
                    #customer_details_dict = _return_customer_minimum_details(customer)
                    customer_details_dict = {
                                        'name': customer.name,
                                        'reference_code': customer.reference_code,
                                        'customer_key'  : customer.key_in_str,
                                        }
                
                return jsonify(customer_details_dict)
                    
            
            return create_api_message(status_code=StatusCode.BAD_REQUEST)
            
    else:
        return create_api_message(status_code=StatusCode.BAD_REQUEST)    
    
@customer_api_bp.route('/customer-key/<customer_key>', methods=['GET'])
@show_request_info
@auth_token_required
@device_is_activated
def read_customer_by_customer_key(customer_key):
    
    logger.debug('customer_key=%s', customer_key)
    
    if is_not_empty(customer_key):
        customer_details_dict = None
        db_client = create_db_client(caller_info="read_customer_by_customer_key")
        
        with db_client.context():
            customer = Customer.fetch(customer_key)
            
            if customer:
                customer_details_dict = return_customer_details(customer)
            
        if customer_details_dict:
            return jsonify(customer_details_dict)
            
            
        else:
            logger.warn('Customer with customer_key (%s) is not found', customer_key)
            return create_api_message(status_code=StatusCode.BAD_REQUEST)
            
    else:
        return create_api_message(status_code=StatusCode.BAD_REQUEST)    


def _update_customer(updating_customer, customer_data_in_json):
    
    is_email_used                           = False
    is_mobile_phone_used                    = False
    customer_is_exist                       = False
    checking_user_acct                      = None
    is_customer_email_changed               = False
    is_customer_mobile_phone_changed        = False
    is_used_email_same_user_acct            = False
    is_used_mobile_phone_same_user_acct     = False
    
    email               = customer_data_in_json.get('email')
    mobile_phone        = customer_data_in_json.get('mobile_phone')
    
    if mobile_phone:
        mobile_phone = mobile_phone.replace(" ", "")
    
    if updating_customer:
        customer_is_exist = True
        is_customer_email_changed           = updating_customer.email!=email
        is_customer_mobile_phone_changed    = updating_customer.mobile_phone!=mobile_phone
    
    if is_customer_email_changed:
        if is_not_empty(email):
            checking_user_acct = User.get_by_email(email)
            
            if checking_user_acct:
                is_email_used = True
                is_used_email_same_user_acct = checking_user_acct.reference_code == updating_customer.reference_code
            
    if is_customer_mobile_phone_changed:
        if is_not_empty(mobile_phone):
            checking_user_acct = User.get_by_mobile_phone(mobile_phone)
        
            if checking_user_acct:
                is_mobile_phone_used = True
                is_used_mobile_phone_same_user_acct = checking_user_acct.reference_code == updating_customer.reference_code
                    
    logger.debug('is_customer_email_changed=%s', is_customer_email_changed)
    logger.debug('is_customer_mobile_phone_changed=%s', is_customer_mobile_phone_changed)
    logger.debug('is_email_used=%s', is_email_used)
    logger.debug('is_mobile_phone_used=%s', is_mobile_phone_used)
    logger.debug('is_used_email_same_user_acct=%s', is_used_email_same_user_acct)
    logger.debug('is_used_mobile_phone_same_user_acct=%s', is_used_mobile_phone_same_user_acct)
    
    if is_email_used==False and is_mobile_phone_used==False:
        logger.debug('Going to update customer details') 
        
        birth_date              = customer_data_in_json.get('birth_date')
            
        if is_not_empty(birth_date):
            birth_date = parse_generic_date(birth_date)
        
        logger.debug('birth_date=%s', birth_date)
        
        Customer.update(customer                = updating_customer, 
                        name                    = customer_data_in_json.get('name'), 
                        email                   = customer_data_in_json.get('email'), 
                        gender                  = customer_data_in_json.get('gender'),
                        birth_date              = birth_date,
                        mobile_phone            = mobile_phone, 
                        password                = customer_data_in_json.get('password'),
                        merchant_reference_code = customer_data_in_json.get('merchant_reference_code')
                        )
    
    
    if customer_is_exist:
        if is_email_used==True:
            return create_api_message('Email have been taken', status_code=StatusCode.BAD_REQUEST)
        
        elif is_mobile_phone_used==True:
            return create_api_message('Mobile phone have been taken', status_code=StatusCode.BAD_REQUEST)
        else:
            return create_api_message(status_code=StatusCode.OK)
    else:
        return create_api_message(gettext('Customer is not exist'), status_code=StatusCode.BAD_REQUEST)
    

#@customer_api_bp.route('/customer-key/<customer_key>', methods=['PUT'])
@customer_api_bp.route('/reference-code/<reference_code>', methods=['PUT'])
@show_request_info
@auth_token_required
@device_is_activated
@request_values
@request_headers
def update_customer_by_reference_code(request_values, request_headers, reference_code):
    customer_data_in_json   = request_values
    updating_customer_form  = CustomerDetailsUpdateForm(ImmutableMultiDict(customer_data_in_json))
    
    logger.debug('update_customer: customer_data_in_json=%s', customer_data_in_json)
    
    try:
        if updating_customer_form.validate():
            
                
            logger.debug('customer update input is valid')
            customer_data_in_json = request_values
            db_client = create_db_client(caller_info="update_customer")
            
            with db_client.context():
                account_id        = request_headers.get('x-acct-id')
                logger.debug('update_customer: account_id=%s', account_id)
                logger.debug('update_customer: reference_code=%s', reference_code)
                
                merchant_acct       = MerchantAcct.fetch(account_id)
                if merchant_acct:
                    updating_customer = Customer.get_by_reference_code(reference_code, merchant_acct)
                else:
                    updating_customer = None
            
            if updating_customer:
                logger.debug('update_customer: updating_customer is not none')
                with db_client.context():
                    return _update_customer(updating_customer, customer_data_in_json)
            else:
                return create_api_message(gettext('Customer is not exist'), status_code=StatusCode.BAD_REQUEST)
                
                
            
        else:
            logger.warn('customer registration input is invalid')
            error_message = updating_customer_form.create_rest_return_error_message()
            
            return create_api_message(error_message, status_code=StatusCode.BAD_REQUEST)
    except:
        logger.error('Fail to update customer due to %s', get_tracelog())
        
        return create_api_message(status_code=StatusCode.BAD_REQUEST)
    
@customer_api_bp.route('/customer-key/<customer_key>', methods=['PUT'])
@show_request_info
@auth_token_required
@device_is_activated
@request_values
def update_customer_by_customer_key(request_values, customer_key):
    customer_data_in_json   = request_values
    updating_customer_form  = CustomerDetailsUpdateForm(ImmutableMultiDict(customer_data_in_json))
    
    logger.debug('update_customer: customer_data_in_json=%s', customer_data_in_json)
    
    try:
        if updating_customer_form.validate():
            
                
            logger.debug('customer update input is valid')
            customer_data_in_json = request_values
            db_client = create_db_client(caller_info="update_customer")
            
            with db_client.context():
                updating_customer   = Customer.fetch(customer_key)
                
                return _update_customer(updating_customer, customer_data_in_json)
            
        else:
            logger.warn('customer registration input is invalid')
            error_message = updating_customer_form.create_rest_return_error_message()
            
            return create_api_message(error_message, status_code=StatusCode.BAD_REQUEST)
    except:
        logger.error('Fail to register customer due to %s', get_tracelog())
        
        return create_api_message(status_code=StatusCode.BAD_REQUEST)    
        
@customer_api_bp.route('/customer-key/<customer_key>', methods=['DELETE'])
@auth_token_required 
@device_is_activated
def delete_customer(customer_key):
    if is_not_empty(customer_key):
        is_found = False
        db_client = create_db_client(caller_info="delete_customer")
        with db_client.context():
            customer = Customer.fetch(customer_key) 
            if customer:
                _delete_customer(customer)
                is_found = True
        
        if is_found:
            return create_api_message(status_code=StatusCode.NO_CONTENT)
        else:
            return create_api_message(gettext('Customer is not exist'), status_code=StatusCode.BAD_REQUEST)
    else:
        return create_api_message(status_code=StatusCode.BAD_REQUEST) 
    
@customer_api_bp.route('/reference-code/<reference_code>', methods=['DELETE'])
@auth_token_required 
@device_is_activated
@request_headers
def delete_customer_by_reference_code(request_headers, reference_code):
    if is_not_empty(reference_code):
        is_found = False
        db_client = create_db_client(caller_info="delete_customer")
        with db_client.context():
            account_id        = request_headers.get('x-acct-id')
            logger.debug('update_customer: account_id=%s', account_id)
            logger.debug('update_customer: reference_code=%s', reference_code)
            
            merchant_acct       = MerchantAcct.fetch(account_id)
            if merchant_acct:
                customer = Customer.get_by_reference_code(reference_code, merchant_acct) 
            
            if customer:
                _delete_customer(customer)
                is_found = True
        
        if is_found:
            return create_api_message(status_code=StatusCode.NO_CONTENT)
        else:
            return create_api_message(gettext('Customer is not exist'), status_code=StatusCode.BAD_REQUEST)
    else:
        return create_api_message(status_code=StatusCode.BAD_REQUEST)            


@model_transactional(desc='Delete customer transaction')
def _delete_customer(customer):
    CustomerMembership.delete_all_by_customer(customer)
    CustomerTierMembership.delete_all_by_customer(customer)
    CustomerTransaction.delete_all_by_customer(customer)
    CustomerPointReward.delete_all_by_customer(customer)
    CustomerStampReward.delete_all_by_customer(customer)
    CustomerPrepaidReward.delete_all_by_customer(customer)
    CustomerEntitledVoucher.delete_all_by_customer(customer)
    RevertedCustomerPointReward.delete_all_by_customer(customer)
    RevertedCustomerStampReward.delete_all_by_customer(customer)
    CustomerEntitledTierRewardSummary.delete_all_by_customer(customer)
    customer.delete()

def __register_user_as_customer(existing_user_acct, outlet):
    created_customer    = None
    merchant_acct       = outlet.merchant_acct_entity
    logger.debug('Valid granted outlet key for merchant acct')
    
    email           = existing_user_acct.email
    mobile_phone    = existing_user_acct.mobile_phone
    
    logger.debug('email=%s', email)
    logger.debug('mobile_phone=%s', mobile_phone)
    
    checking_customer = Customer.get_by_email(email, merchant_acct=merchant_acct) 
    
    if checking_customer:
        is_email_used = True
    else:
        if is_not_empty(mobile_phone):
            checking_customer = Customer.get_by_mobile_phone(mobile_phone, merchant_acct=merchant_acct)
            if checking_customer:
                is_mobile_phone_used = True
    
    logger.debug('is_email_used=%s', is_email_used)
    logger.debug('is_mobile_phone_used=%s', is_mobile_phone_used)
    
    if is_email_used == False and is_mobile_phone_used == False:
    
        created_customer = Customer.create_from_user(existing_user_acct, outlet=outlet)
        
        create_merchant_registered_customer_upstream_for_merchant(created_customer)
        create_registered_customer_upstream_for_system(created_customer)
    
        
    logger.debug('created_customer=%s', created_customer)
    
    return created_customer
        
            

@customer_api_bp.route('/register-as-customer', methods=['POST'])
@auth_token_required
@device_is_activated
@request_json
@request_headers
def register_user_as_customer(user_data_in_json, request_headers):
    reference_code              = user_data_in_json.get('reference_code')
    
    logger.debug('register_user_as_customer: user_data_in_json=%s', user_data_in_json)
    
    try:
        if is_not_empty(reference_code):
            logger.debug('customer registration input is valid')
            db_client = create_db_client(caller_info="register_user_as_customer")
            
            created_customer        = None
            existing_user_acct      = None
            is_email_used           = False
            is_mobile_phone_used    = False
            merchant_act_key        = None
            
            merchant_acct           = None
            
            outlet_key  = request_headers.get('x-outlet-key')
            acct_id     = request_headers.get('x-acct-id')
            
            logger.debug('register_user_as_customer: outlet_key=%s', outlet_key)
            logger.debug('register_user_as_customer: acct_id=%s', acct_id)
            logger.debug('register_user_as_customer: reference_code=%s', reference_code)
            
            with db_client.context():
                existing_user_acct  = User.get_by_reference_code(reference_code)
                
                logger.debug('register_user_as_customer: existing_user_acct=%s', existing_user_acct)
                
                if existing_user_acct:
                    
                    outlet              = Outlet.fetch(outlet_key)
                        
                    if outlet:
                        merchant_acct       = outlet.merchant_acct_entity
                        logger.debug('Valid granted outlet key for merchant acct')
                        
                        created_customer = Customer.get_by_reference_code(reference_code, merchant_acct)
                         
                        logger.debug('register_user_as_customer: created_customer=%s', created_customer) 
                         
                        if created_customer is None:
                            
                            created_customer = __register_user_as_customer(existing_user_acct, outlet)
                            

                        logger.debug('created_customer=%s', created_customer)
                        
                    else:
                        logger.warn('Invalid granted outlet key or merchant account id')
                
                if created_customer:
                    
                    
                    '''
                    response_data = {
                                    'customer_key'              : created_customer.key_in_str,
                                    'registered_datetime'       : created_customer.registered_datetime.strftime("%d-%m-%Y %H:%M:%S"),
                                    'reference_code'            : created_customer.reference_code,
                                    'merchant_account_key'      : merchant_act_key,
                                    'company_name'              : merchant_acct.company_name,
                                    'outlet_key'                : outlet_key,  
                                    #'user_details'              : user_details_dict(existing_user_acct),
                                    }
                    
                    response_data.update(user_details_dict(existing_user_acct))
                    '''
                    response_data = return_customer_details(created_customer)
                    
                    logger.debug('register_user_as_customer debug: response_data=%s', response_data)
                    
                    return create_api_message(status_code=StatusCode.OK, **response_data)
                    
                else:
                    if is_email_used==True:
                        return create_api_message('Email have been taken', status_code=StatusCode.BAD_REQUEST)
                    
                    elif is_mobile_phone_used==True:
                        return create_api_message('Mobile phone have been taken', status_code=StatusCode.BAD_REQUEST)
                    else:
                        return create_api_message('Failed to register customer', status_code=StatusCode.BAD_REQUEST)
            
        else:
            logger.warn('customer registration input is invalid')
            
            return create_api_message("Missing register customer input data", status_code=StatusCode.BAD_REQUEST)
    except:
        logger.error('Fail to register customer due to %s', get_tracelog())
        
        return create_api_message(status_code=StatusCode.BAD_REQUEST)

@customer_api_bp.route('/search/<limit>', methods=['POST'])
@auth_token_required
@device_is_activated
@request_json
@request_headers
#@test_session_expired
def search_customer(search_member_data_in_json, request_headers, limit):
    
    #search_member_data_in_json   = request.get_json()
    search_customer_form         = CustomerSearchForm(ImmutableMultiDict(search_member_data_in_json))
    
    logger.debug('search_member_data_in_json=%s', search_member_data_in_json)
    
    db_client = create_db_client(caller_info="search_customer")
    customer_list = []
    
    if search_customer_form.validate():
        name                        = search_customer_form.name.data
        mobile_phone                = search_customer_form.mobile_phone.data
        email                       = search_customer_form.email.data
        reference_code              = search_customer_form.reference_code.data
        merchant_reference_code     = search_customer_form.merchant_reference_code.data
        #customer_data               = search_customer_form.customer_data.data
    
        acct_id     = request_headers.get('x-acct-id')
        
        logger.debug('acct_id=%s', acct_id)
        logger.debug('name=%s', name)
        logger.debug('email=%s', email)
        logger.debug('mobile_phone=%s', mobile_phone)
        logger.debug('reference_code=%s', reference_code)
        logger.debug('merchant_reference_code=%s', merchant_reference_code)
        
        limit_int = int(limit)
        
        with db_client.context():
            merchant_acct   = MerchantAcct.fetch(acct_id)
            (search_results, total_count, next_cursor)  = Customer.search_merchant_customer(merchant_acct, 
                                                                                            name                    = name, 
                                                                                            email                   = email, 
                                                                                            mobile_phone            = mobile_phone, 
                                                                                            reference_code          = reference_code,
                                                                                            merchant_reference_code = merchant_reference_code,
                                                                                            limit                   = limit_int,
                                                                                            )
            
            logger.debug('total_count=%s', total_count)
            logger.debug('next_cursor=%s', next_cursor)
            
            for c in search_results:
                user_details = c.registered_user_acct
                customer_details_dict = c.to_dict(
                                            date_format="%d-%m-%Y", 
                                            datetime_format="%d-%m-%Y %H:%M:%S",
                                            excluded_dict_properties=['registered_merchant_acct'],
                                            )
                customer_details_dict['customer_key']               = c.key_in_str
                customer_details_dict['is_email_verified']          = user_details.is_email_verified
                customer_details_dict['is_mobile_phone_verified']   = user_details.is_mobile_phone_verified
                
                #if 'reward' in customer_data:
                
                if 'entitled_voucher_summary' in customer_details_dict:
                    customer_details_dict['voucher_summary']            = c.entitled_voucher_summary
                    #del customer_details_dict['entitled_voucher_summary'] 
                
                
                logger.debug('customer_details_dict=%s', customer_details_dict)
                
                #if 'membership' in customer_data
                customer_basic_memberships_list = _list_customer_basic_memberships(c)
                if customer_basic_memberships_list:
                    customer_details_dict['basic_memberships'] = customer_basic_memberships_list
                
                tier_membership_data = _get_tier_membership(c)
                if tier_membership_data:
                    customer_details_dict['tier_membership'] = tier_membership_data
                
                customer_list.append(customer_details_dict)
    
        #return create_api_message(status_code=customer_list.OK, customer_list=customer_list)
        customer_details_list_json =  jsonify(customer_list)
            
            
        logger.debug('customer_details_list_json=%s', customer_details_list_json)
            
        return customer_details_list_json
    else:
        error_message = search_customer_form.create_rest_return_error_message()
        return create_api_message(error_message, status_code=StatusCode.BAD_REQUEST)
        #return create_api_message(gettext('Invalid input'), status_code=StatusCode.BAD_REQUEST)
        #return create_api_message('Invalid input', status_code=StatusCode.BAD_REQUEST)

@customer_api_bp.route('/search/customer/<limit>', methods=['POST'])
@auth_token_required
@device_is_activated
@request_json
@request_headers
#@test_session_expired
def search_customer_details_only(search_member_data_in_json, request_headers, limit):
    
    #search_member_data_in_json   = request.get_json()
    search_customer_form         = CustomerSearchForm(ImmutableMultiDict(search_member_data_in_json))
    
    logger.debug('search_member_data_in_json=%s', search_member_data_in_json)
    
    db_client = create_db_client(caller_info="search_customer_details_only")
    customer_list = []
    
    if search_customer_form.validate():
        name                        = search_customer_form.name.data
        mobile_phone                = search_customer_form.mobile_phone.data
        email                       = search_customer_form.email.data
        reference_code              = search_customer_form.reference_code.data
        merchant_reference_code     = search_customer_form.merchant_reference_code.data
        #customer_data               = search_customer_form.customer_data.data
    
        acct_id     = request_headers.get('x-acct-id')
        
        logger.debug('acct_id=%s', acct_id)
        logger.debug('name=%s', name)
        logger.debug('email=%s', email)
        logger.debug('mobile_phone=%s', mobile_phone)
        logger.debug('reference_code=%s', reference_code)
        logger.debug('merchant_reference_code=%s', merchant_reference_code)
        
        limit_int = int(limit)
        
        with db_client.context():
            merchant_acct   = MerchantAcct.fetch(acct_id)
            (search_results, total_count, next_cursor)  = Customer.search_merchant_customer(merchant_acct, 
                                                                                            name                    = name, 
                                                                                            email                   = email, 
                                                                                            mobile_phone            = mobile_phone, 
                                                                                            reference_code          = reference_code,
                                                                                            merchant_reference_code = merchant_reference_code,
                                                                                            limit                   = limit_int,
                                                                                            )
            
            logger.debug('total_count=%s', total_count)
            logger.debug('next_cursor=%s', next_cursor)
            
            for c in search_results:
                user_details = c.registered_user_acct
                customer_details_dict = c.to_dict(
                                            date_format="%d-%m-%Y", 
                                            datetime_format="%d-%m-%Y %H:%M:%S",
                                            excluded_dict_properties=[
                                                'registered_merchant_acct', 'tier_membership', 'memberships_list',
                                                'reward_summary', 'prepaid_summary',
                                                'entitled_voucher_summary', 'entitled_birthday_reward_summary',
                                                'entitled_membership_reward_summary', 'entitled_lucky_draw_ticket_summary',
                                                'kpi_summary','device_details',
                                                ],
                                            )
                customer_details_dict['customer_key']               = c.key_in_str
                customer_details_dict['is_email_verified']          = user_details.is_email_verified
                customer_details_dict['is_mobile_phone_verified']   = user_details.is_mobile_phone_verified
                
                logger.debug('customer_details_dict=%s', customer_details_dict)
                
                customer_list.append(customer_details_dict)
    
        #return create_api_message(status_code=customer_list.OK, customer_list=customer_list)
        customer_details_list_json =  jsonify(customer_list)
            
            
        logger.debug('customer_details_list_json=%s', customer_details_list_json)
            
        return customer_details_list_json
    else:
        error_message = search_customer_form.create_rest_return_error_message()
        return create_api_message(error_message, status_code=StatusCode.BAD_REQUEST)

@customer_api_bp.route('/search/customer-for-membership/<limit>', methods=['POST'])
@auth_token_required
@device_is_activated
@request_json
@request_headers
#@test_session_expired
def search_customer_for_membership(search_member_data_in_json, request_headers, limit):
    
    #search_member_data_in_json   = request.get_json()
    search_customer_form         = CustomerSearchForm(ImmutableMultiDict(search_member_data_in_json))
    
    logger.debug('search_member_data_in_json=%s', search_member_data_in_json)
    
    db_client = create_db_client(caller_info="search_customer")
    customer_list = []
    
    if search_customer_form.validate():
        name                        = search_customer_form.name.data
        mobile_phone                = search_customer_form.mobile_phone.data
        email                       = search_customer_form.email.data
        reference_code              = search_customer_form.reference_code.data
        merchant_reference_code     = search_customer_form.merchant_reference_code.data
        #customer_data               = search_customer_form.customer_data.data
    
        acct_id     = request_headers.get('x-acct-id')
        
        logger.debug('acct_id=%s', acct_id)
        logger.debug('name=%s', name)
        logger.debug('email=%s', email)
        logger.debug('mobile_phone=%s', mobile_phone)
        logger.debug('reference_code=%s', reference_code)
        logger.debug('merchant_reference_code=%s', merchant_reference_code)
        
        limit_int = int(limit)
        
        with db_client.context():
            merchant_acct   = MerchantAcct.fetch(acct_id)
            (search_results, total_count, next_cursor)  = Customer.search_merchant_customer(merchant_acct, 
                                                                                            name                    = name, 
                                                                                            email                   = email, 
                                                                                            mobile_phone            = mobile_phone, 
                                                                                            reference_code          = reference_code,
                                                                                            merchant_reference_code = merchant_reference_code,
                                                                                            limit                   = limit_int,
                                                                                            )
            
            logger.debug('total_count=%s', total_count)
            logger.debug('next_cursor=%s', next_cursor)
            
            for c in search_results:
                customer_details_dict = c.to_dict(
                                            date_format="%d-%m-%Y", 
                                            datetime_format="%d-%m-%Y %H:%M:%S",
                                            excluded_dict_properties=[
                                                'registered_merchant_acct', 'tier_membership', 'memberships_list',
                                                'reward_summary', 'prepaid_summary',
                                                'entitled_voucher_summary', 'entitled_birthday_reward_summary',
                                                'entitled_membership_reward_summary', 'entitled_lucky_draw_ticket_summary',
                                                'kpi_summary','device_details',
                                                ],
                                            )
                customer_details_dict['customer_key']               = c.key_in_str
                
                logger.debug('customer_details_dict=%s', customer_details_dict)
                
                #if 'membership' in customer_data
                customer_basic_memberships_list = _list_customer_basic_memberships(c)
                if customer_basic_memberships_list:
                    customer_details_dict['basic_memberships'] = customer_basic_memberships_list
                
                tier_membership_data = _get_tier_membership(c)
                if tier_membership_data:
                    customer_details_dict['tier_membership'] = tier_membership_data
                
                customer_list.append(customer_details_dict)
    
        #return create_api_message(status_code=customer_list.OK, customer_list=customer_list)
        customer_details_list_json =  jsonify(customer_list)
            
            
        logger.debug('customer_details_list_json=%s', customer_details_list_json)
            
        return customer_details_list_json
    else:
        error_message = search_customer_form.create_rest_return_error_message()
        return create_api_message(error_message, status_code=StatusCode.BAD_REQUEST)



def _list_customer_basic_memberships(customer):
    customer_membership_final_list = []
    
    if is_not_empty(customer.memberships_list):
        customer_memberships_list = CustomerMembership.list_all_by_customer(customer)
        if is_not_empty(customer_memberships_list):
            merchant_acct = customer.registered_merchant_acct
            merchant_memberships_list = MerchantMembership.list_by_merchant_acct(merchant_acct)
            
            for cm in customer_memberships_list:
                for mm in merchant_memberships_list:
                    logger.debug('cm=%s', cm)
                    if mm.key_in_str == cm.merchant_membership_key:
                        membership_data = {
                                                        'key'                   : mm.key_in_str,
                                                        'label'                 : mm.label,
                                                        'card_image'            : mm.membership_card_image,
                                                        'desc'                  : mm.desc if is_not_empty(mm.desc) else '',
                                                        'terms_and_conditions'  : mm.terms_and_conditions if is_not_empty(mm.terms_and_conditions) else '',
                                                        'is_tier'               : False,
                                                        'entitled_date'         : cm.entitled_date.strftime('%d-%m-%Y'),
                                                        'expiry_date'           : cm.expiry_date.strftime('%d-%m-%Y'),
                                                        
                                                        }
                        
                        if cm.renewed_date is not None:
                            membership_data['renewed_date'] = cm.renewed_date.strftime('%d-%m-%Y'),
                        
                        customer_membership_final_list.append(membership_data)
                        break
    return customer_membership_final_list

def _get_tier_membership(customer):
            
    if is_not_empty(customer.tier_membership):
        merchant_tier_membership = customer.tier_membership_entity
        customer_tier_membership = CustomerTierMembership.get_by_customer_and_merchant_tier_membership(customer, merchant_tier_membership)
        
        membership_data = {
                        'key'                   : customer_tier_membership.merchant_tier_membership_key,
                        'label'                 : merchant_tier_membership.label,
                        'card_image'            : merchant_tier_membership.membership_card_image,
                        'desc'                  : merchant_tier_membership.desc if is_not_empty(merchant_tier_membership.desc) else '',
                        'terms_and_conditions'  : merchant_tier_membership.terms_and_conditions if is_not_empty(merchant_tier_membership.terms_and_conditions) else '',
                        'entitled_date'         : customer_tier_membership.entitled_date.strftime('%d-%m-%Y'),
                        'expiry_date'           : customer_tier_membership.expiry_date.strftime('%d-%m-%Y'),
                        'is_tier'               : True,
                        }
        
        return membership_data
        
@customer_api_bp.route('/customer-key/<customer_key>/customer-membership', methods=['GET'])
@auth_token_required
def read_customer_membership(customer_key):
    db_client = create_db_client(caller_info="read_customer_membership")
    
    with db_client.context():
        customer        = Customer.fetch(customer_key)
        customer_membershis_list = _list_customer_basic_memberships(customer)
        tier_membership_data = _get_tier_membership(customer)
        if tier_membership_data:
            customer_membershis_list.append(tier_membership_data)

    return jsonify(customer_membershis_list)  

@customer_api_bp.route('/reference-code/<reference_code>/list-membership', methods=['GET'])
@auth_token_required
@request_headers
#@test_session_expired
def list_customer_membership(request_headers, reference_code):
    if is_not_empty(reference_code):
        acct_id   = request_headers.get('x-acct-id')
        db_client = create_db_client(caller_info="list_customer_membership")
        customer_membershis_list = []
        
        logger.debug('reference_code=%s', reference_code)
        
        with db_client.context():
            merchant_acct   = MerchantAcct.fetch(acct_id)
            customer        = Customer.get_by_reference_code(reference_code, merchant_acct)
            
            customer_membershis_list = _list_customer_basic_memberships(customer)
            tier_membership_data = _get_tier_membership(customer)
            if tier_membership_data:
                customer_membershis_list.append(tier_membership_data)
    
        return jsonify(customer_membershis_list)
    else:
        return create_api_message(status_code=StatusCode.BAD_REQUEST)    
                


@customer_api_bp.route('/reference-code/<reference_code>/transaction/limit/<limit>', methods=['GET'])
@auth_token_required
@request_headers
#@test_session_expired
def list_customer_transaction(request_headers, reference_code, limit):
    
    logger.debug('reference_code=%s', reference_code)
    
    if is_not_empty(reference_code):
        limit_int = int(limit, 10)
        acct_id   = request_headers.get('x-acct-id')
        db_client = create_db_client(caller_info="read_customer_sales_transaction")
        transactions_list = []
        with db_client.context():
            merchant_acct = MerchantAcct.fetch(acct_id)
            customer = Customer.get_by_reference_code(reference_code, merchant_acct)
            
        
        if customer:
            dict_properties  = ['transaction_id', 'invoice_id', 'remarks', 'tax_amount', 'transact_amount', 'reward_giveaway_method',
                               'entitled_reward_summary', 'entitled_voucher_summary', 'entitled_prepaid_summary', 
                               #'transact_outlet_details', 
                               'transact_datetime', 'created_datetime',  'transact_outlet_key', 'is_revert', 'reverted_datetime',
                               'transact_by_username', 'is_reward_redeemed', 'is_sales_transaction', 'allow_to_revert',
                               'is_membership_purchase', 'purchased_merchant_membership_key', 'is_membership_renew',
                               ]
            with db_client.context():
                result       = CustomerTransaction.list_customer_transaction(customer, limit=limit_int)
                for r in result:
                    transactions_list.append(r.to_dict(dict_properties=dict_properties,  date_format="%d-%m-%Y", datetime_format="%d-%m-%Y %H:%M:%S"))
            
            return jsonify(transactions_list)
            
        else:
            logger.warn('Customer transaction with reference code (%s) is not found', reference_code)
            return create_api_message(status_code=StatusCode.BAD_REQUEST)
            
    else:
        return create_api_message(status_code=StatusCode.BAD_REQUEST)
    
@customer_api_bp.route('/reference-code/<reference_code>/transaction/transaction_id/<transaction_id>/revert', methods=['POST'])
@auth_token_required
def revert_customer_transaction(reference_code, transaction_id):
    
    logger.debug('transaction_id=%s', transaction_id)
    
    if is_not_empty(reference_code) and is_not_empty(transaction_id):
        #acct_id   = request_headers.get('x-acct-id')
        db_client = create_db_client(caller_info="revert_customer_sales_transaction")
        
        with db_client.context():
            customer_transactionn       = CustomerTransaction.get_by_transaction_id(transaction_id);
        
        if customer_transactionn:
            with db_client.context():
                merchant_username       = get_logged_in_api_username()
                reverted_by             = MerchantUser.get_by_username(merchant_username)
                
                reverted_datetime_utc   = datetime.utcnow()
                __revert_customer_transaction(customer_transactionn, reverted_by, reverted_datetime=reverted_datetime_utc)
            
            return create_api_message(status_code=StatusCode.OK, reverted_datetime = customer_transactionn.reverted_datetime.strftime('%d-%m-%Y %H:%M:%S'))
        else:    
            return create_api_message(gettext('Failed to find transaction'), status_code=StatusCode.BAD_REQUEST)
        
        
            
    else:
        return create_api_message(gettext('Missing reference code or transaction id'), status_code=StatusCode.BAD_REQUEST)    

@model_transactional(desc="revert_customer_transaction")
def __revert_customer_transaction(customer_transction, reverted_by, reverted_datetime):     
    return revert_transaction(customer_transction, reverted_by, reverted_datetime=reverted_datetime)

    
@customer_api_bp.route('/reference-code/<reference_code>/redemption/limit/<limit>', methods=['GET'])
@auth_token_required
@request_headers
#@test_session_expired
def list_customer_redemption(request_headers, reference_code, limit):
    
    logger.debug('reference_code=%s', reference_code)
    
    if is_not_empty(reference_code):
        limit_int = int(limit, 10)
        acct_id   = request_headers.get('x-acct-id')
        db_client = create_db_client(caller_info="read_customer_sales_transaction")
        redemptions_list = []
        with db_client.context():
            merchant_acct = MerchantAcct.fetch(acct_id)
            customer = Customer.get_by_reference_code(reference_code, merchant_acct)
            
        
        if customer:
            dict_properties  = ['transaction_id', 'invoice_id', 'remarks', 'reward_format', 'redeemed_amount', 'redeemed_summary',
                                   'redeemed_datetime', 'is_revert', 'reverted_datetime',
                                   'redeemed_by_username', 
                                   ]
            with db_client.context():
                result       = CustomerRedemption.list_customer_redemption(customer, limit=limit_int)
                for customer_redemption in result:
                    #fixed reward int and float convertion issue
                    _resolve_int_and_float_for_redeem_summary(customer_redemption.redeemed_summary)
                    customer_redemption_dict = customer_redemption.to_dict(dict_properties=dict_properties,  date_format="%d-%m-%Y", datetime_format="%d-%m-%Y %H:%M:%S")
                    redemptions_list.append(customer_redemption_dict)
            
            return jsonify(redemptions_list)
            
        else:
            logger.warn('Customer transaction with reference code (%s) is not found', reference_code)
            return create_api_message(status_code=StatusCode.BAD_REQUEST)
            
    else:
        return create_api_message(status_code=StatusCode.BAD_REQUEST)    

def _resolve_int_and_float_for_redeem_summary(redeem_summary):
    if 'prepaid' in redeem_summary:
        redeem_summary['prepaid']['amount'] = float(redeem_summary['prepaid']['amount'])
    elif 'point' in redeem_summary:
        redeem_summary['point']['amount'] = float(redeem_summary['point']['amount'])

    return redeem_summary

    
@customer_api_bp.route('/reference-code/<reference_code>/redemption/transaction_id/<transaction_id>/revert', methods=['POST'])
@auth_token_required
@request_headers
def revert_customer_redemption(request_headers, reference_code, transaction_id):
    
    logger.debug('transaction_id=%s', transaction_id)
    
    if is_not_empty(reference_code) and is_not_empty(transaction_id):
        #acct_id   = request_headers.get('x-acct-id')
        db_client = create_db_client(caller_info="revert_customer_sales_transaction")
        
        with db_client.context():
            customer_redemption    = CustomerRedemption.get_by_transaction_id(transaction_id);
        
        if customer_redemption:
            with db_client.context():
                merchant_username       = get_logged_in_api_username()
                reverted_by             = MerchantUser.get_by_username(merchant_username)
                
                reverted_datetime_utc   = datetime.utcnow()
                __revert_customer_redemption(customer_redemption, reverted_by, reverted_datetime=reverted_datetime_utc)
            
            return create_api_message(status_code=StatusCode.OK, reverted_datetime = customer_redemption.reverted_datetime.strftime('%d-%m-%Y %H:%M:%S'))
        else:    
            return create_api_message(gettext('Failed to find transaction'), status_code=StatusCode.BAD_REQUEST)
        
        
            
    else:
        return create_api_message(gettext('Missing reference code or transaction id'), status_code=StatusCode.BAD_REQUEST)
    
@model_transactional(desc="revert_customer_redemption")
def __revert_customer_redemption(customer_redemption, reverted_by, reverted_datetime=None):     
       
    return revert_redemption(customer_redemption, reverted_by, reverted_datetime=reverted_datetime)


          


import setuptools
with open("README.md", "r") as fh:
    long_description = fh.read()
    
setuptools.setup(
     name='trex_apis',  
     version='1.7.5',
     author="Jack Lok",
     author_email="sglok77@gmail.com",
     description="TRex APIs package",
     long_description=long_description,
     long_description_content_type="text/markdown",
     url="https://bitbucket.org/lokjac/trex-apis",
     packages=setuptools.find_packages(),
     classifiers=[
         "Programming Language :: Python :: 3",
         "License :: OSI Approved :: MIT License",
         "Operating System :: OS Independent",
     ],
     install_requires=[            
          'cryptography',
          'basicauth',
          'trex_conf',
          'trex_model',
          'trex_lib',
          'trex_program',
      ],
 )




'''
Created on 1 Jul 2021

@author: jacklok
'''
from functools import wraps
from flask import request, session, abort
from trexlib.utils.string_util import is_not_empty
from trexlib.utils.crypto_util import decrypt_json
import logging
from datetime import datetime
from trexlib.utils.log_util import get_tracelog
from trexmodel.models.datastore.loyalty_models import LoyaltyDeviceSetting
from trexmodel.models.datastore.pos_models import POSSetting
from trexmodel.utils.model.model_util import create_db_client
from trexmodel.models.datastore.user_models import User
from trexapi.conf import API_ERR_CODE_INVALID_SESSION,\
    API_ERR_CODE_EXPIRED_SESSION, API_ERR_CODE_DUPLICATED_SESSION
from trexapi.utils.api_helpers import create_api_message, StatusCode
from trexlib.libs.google.gae_app_logger import AppLogger

#logger = logging.getLogger('decorator')
logger = AppLogger.get_logger('target_debug')

def test_session_expired(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        return ("Authenticated token is expired", 401)
    
    return decorated_function

def auth_token_required(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        auth_token  = request.headers.get('x-auth-token')
        acct_id     = request.headers.get('x-acct-id')
            
        logger.debug('auth_token=%s', auth_token)
        logger.debug('acct_id=%s', acct_id)
        
        if is_not_empty(auth_token):
            if 'bypass' == auth_token:
                return f(*args, **kwargs)
            else:
                try:
                    auth_details_json = decrypt_json(auth_token)
                      
                except:
                    logger.error('Authenticated token is not valid')
                    return ("Authenticated token is not valid", 401)
                
                logger.debug('auth_details_json=%s', auth_details_json)
                
                if auth_details_json:
                    expiry_datetime     = auth_details_json.get('expiry_datetime')
                    acct_id_from_token  = auth_details_json.get('acct_id')
                    
                    logger.debug('acct_id from decrypted token=%s', acct_id_from_token)
                    logger.debug('expiry_datetime from decrypted token=%s', expiry_datetime)
                    
                    if is_not_empty(expiry_datetime) and is_not_empty(acct_id_from_token) and acct_id==acct_id_from_token:
                        expiry_datetime = datetime.strptime(expiry_datetime, '%d-%m-%Y %H:%M:%S')
                        logger.debug('expiry_datetime=%s', expiry_datetime)
                        
                        now             = datetime.now()
                        if now < expiry_datetime: 
                            logger.debug('auth token is still valid')
                            return f(*args, **kwargs)
                        else:
                            logger.debug('auth token is not logger valid')
                            
                            return create_api_message('Authenticated token is expired', status_code=StatusCode.UNAUTHORIZED,)
                    else:
                        return create_api_message('Authenticated token is invalid', status_code=StatusCode.UNAUTHORIZED,)
        
        return create_api_message('Authenticated token is required', status_code=StatusCode.UNAUTHORIZED,)

    return decorated_function

def show_request_info(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        logger.info('----------------------------->>>>> request url=%s, method=%s, function=%s', request.url, request.method, f.__name__)
        return f(*args, **kwargs)
    
    return decorated_function

      
def user_auth_token_required_and_check_duplicated_session(check_duplicated_session=True):
    
    def decorator(f):
        return _user_auth_token_required_decorated_function(f, check_duplicated_session=check_duplicated_session, pass_reference_code=True)
        '''
        @wraps(f)
        def decorated_function(*args, **kwargs):
            user_auth_token     = request.headers.get('x-auth-token')
            user_reference_code = request.headers.get('x-reference-code')
            
            logger.debug('user_auth_token=%s', user_auth_token)
            logger.debug('user_reference_code=%s', user_reference_code)
            logger.debug('check_duplicated_session=%s', check_duplicated_session)
            
            logger.debug('args=%s', args)
            logger.debug('kwargs=%s', kwargs)
            
            if is_not_empty(user_auth_token):
                logger.debug('user_auth_token is not empty, going to decrypt it')
                try:
                    auth_details_json   = decrypt_json(user_auth_token)
                    logger.debug('auth_details_json=%s', auth_details_json)
                    
                except:
                    logger.error('Failed due to %s', get_tracelog())
                    return create_api_message('Authenticated token is not valid', status_code=StatusCode.UNAUTHORIZED,)
                
                
                
                if auth_details_json:
                    expiry_datetime                 = auth_details_json.get('expiry_datetime')
                    user_reference_code_from_token  = auth_details_json.get('reference_code')
                    decrypted_device_id             = auth_details_json.get('device_id')
                    
                    
                    if is_not_empty(expiry_datetime) and is_not_empty(user_reference_code_from_token) and user_reference_code_from_token == user_reference_code:
                        
                        expiry_datetime = datetime.strptime(expiry_datetime, '%d-%m-%Y %H:%M:%S')
                        logger.debug('expiry_datetime=%s', expiry_datetime)
                        
                        now             = datetime.now()
                        if now < expiry_datetime: 
                            logger.debug('auth token is still valid')
                            logger.debug('new args=%s', (user_reference_code, *args))
                            logger.debug('new kwargs=%s', kwargs)
                            
                            
                            db_client = create_db_client(caller_info="user_auth_token_required_and_check_duplicated_session")
                            
                            
                            with db_client.context():
                                user_acct           = User.get_by_reference_code(user_reference_code_from_token)
                                signin_device_id    = user_acct.signin_device_id
                            
                            if decrypted_device_id==signin_device_id:
                            
                                return f(*(user_reference_code, *args), **kwargs)
                            else:
                                logger.debug('auth token is not logger valid, due to request device id is not same as latest device id')
                            
                                return create_api_message('Authenticated token is not longer valid', status_code=StatusCode.UNAUTHORIZED, error_code=API_ERR_CODE_DUPLICATED_SESSION)
                            
                            
                            
                        else:
                            logger.debug('auth token is not logger valid')
                            
                            #return ("Authenticated token is expired", 401)
                            return create_api_message('Authenticated token is expired', status_code=StatusCode.UNAUTHORIZED, error_code=API_ERR_CODE_EXPIRED_SESSION)
                else:
                    #return ("Authenticated token is invalid", 401)    
                    return create_api_message('Authenticated token is invalid', status_code=StatusCode.UNAUTHORIZED, error_code=API_ERR_CODE_INVALID_SESSION)
            
            #return ("Authenticated token is required", 401)
            return create_api_message('Authenticated token is required', status_code=StatusCode.UNAUTHORIZED, )
            
        return decorated_function
        '''
    return decorator
    
def _user_auth_token_required_decorated_function(f, check_duplicated_session=False, pass_reference_code=False):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        user_auth_token     = request.headers.get('x-auth-token')
        user_reference_code = request.headers.get('x-reference-code')
        
        
        logger.debug('user_auth_token=%s', user_auth_token)
        logger.debug('user_reference_code=%s', user_reference_code)
        logger.debug('check_duplicated_session=%s', check_duplicated_session)
        
        if pass_reference_code:
            kwargs['reference_code'] = user_reference_code
        
        logger.debug('args=%s', args)
        logger.debug('kwargs=%s', kwargs)
        
        
        
        if is_not_empty(user_auth_token):
            if user_auth_token == 'bypass':
                return f(*args, **kwargs)
            
            logger.debug('user_auth_token is not empty, going to decrypt it')
            try:
                auth_details_json   = decrypt_json(user_auth_token)
                logger.debug('auth_details_json=%s', auth_details_json)
                
            except:
                logger.error('Failed due to %s', get_tracelog())
                return create_api_message('Authenticated token is not valid', status_code=StatusCode.UNAUTHORIZED,)
            
            
            
            if auth_details_json:
                expiry_datetime                 = auth_details_json.get('expiry_datetime')
                user_reference_code_from_token  = auth_details_json.get('reference_code')
                decrypted_device_id             = auth_details_json.get('device_id')
                
                
                if is_not_empty(expiry_datetime) and is_not_empty(user_reference_code_from_token) and user_reference_code_from_token == user_reference_code:
                    
                    expiry_datetime = datetime.strptime(expiry_datetime, '%d-%m-%Y %H:%M:%S')
                    logger.debug('expiry_datetime=%s', expiry_datetime)
                    
                    now             = datetime.now()
                    if now < expiry_datetime: 
                        logger.debug('auth token is still valid')
                        logger.debug('new args=%s', (user_reference_code, *args))
                        logger.debug('new kwargs=%s', kwargs)
                        
                        
                        
                        if check_duplicated_session:
                            db_client = create_db_client(caller_info="user_auth_token_required")
                            
                            
                            with db_client.context():
                                user_acct           = User.get_by_reference_code(user_reference_code_from_token)
                                signin_device_id    = user_acct.signin_device_id
                            
                            if decrypted_device_id==signin_device_id:
                            
                                return f(*args, **kwargs)
                                #return f(*(user_reference_code, *args), **kwargs)
                                #return f(user_reference_code, *args, **kwargs)
                            else:
                                logger.debug('auth token is not logger valid, due to request device id is not same as latest device id')
                            
                                return create_api_message('Authenticated token is not longer valid', status_code=StatusCode.UNAUTHORIZED, error_code=API_ERR_CODE_DUPLICATED_SESSION)
                        
                        else:
                            #return f(*(user_reference_code, *args), **kwargs)
                            #return f(user_reference_code, *args, **kwargs)
                            return f(*args, **kwargs)
                        
                    else:
                        logger.debug('auth token is not logger valid')
                        
                        #return ("Authenticated token is expired", 401)
                        return create_api_message('Authenticated token is expired', status_code=StatusCode.UNAUTHORIZED, error_code=API_ERR_CODE_EXPIRED_SESSION)
            else:
                #return ("Authenticated token is invalid", 401)    
                return create_api_message('Authenticated token is invalid', status_code=StatusCode.UNAUTHORIZED, error_code=API_ERR_CODE_INVALID_SESSION)
        
        #return ("Authenticated token is required", 401)
        return create_api_message('Authenticated token is required', status_code=StatusCode.UNAUTHORIZED, )

    return decorated_function

def user_auth_token_required_pass_reference_code(f):
    logger.debug("Calling function: %s", f.__name__)
    return _user_auth_token_required_decorated_function(f, check_duplicated_session=False, pass_reference_code=True)

def user_auth_token_required(f):
    logger.debug("Calling function: %s", f.__name__)
    return _user_auth_token_required_decorated_function(f, check_duplicated_session=False, pass_reference_code=False)
    '''
    @wraps(f)
    def decorated_function(*args, **kwargs):
        user_auth_token     = request.headers.get('x-auth-token')
        user_reference_code = request.headers.get('x-reference-code')
        check_duplicated_session = False
        if kwargs:
            check_duplicated_session = kwargs.get('check_duplicated_session')
        
        logger.debug('user_auth_token=%s', user_auth_token)
        logger.debug('user_reference_code=%s', user_reference_code)
        logger.debug('check_duplicated_session=%s', check_duplicated_session)
        
        logger.debug('args=%s', args)
        logger.debug('kwargs=%s', kwargs)
        
        if is_not_empty(user_auth_token):
            logger.debug('user_auth_token is not empty, going to decrypt it')
            try:
                auth_details_json   = decrypt_json(user_auth_token)
                logger.debug('auth_details_json=%s', auth_details_json)
                
            except:
                logger.error('Failed due to %s', get_tracelog())
                return create_api_message('Authenticated token is not valid', status_code=StatusCode.UNAUTHORIZED,)
            
            
            
            if auth_details_json:
                expiry_datetime                 = auth_details_json.get('expiry_datetime')
                user_reference_code_from_token  = auth_details_json.get('reference_code')
                decrypted_device_id             = auth_details_json.get('device_id')
                
                
                if is_not_empty(expiry_datetime) and is_not_empty(user_reference_code_from_token) and user_reference_code_from_token == user_reference_code:
                    
                    expiry_datetime = datetime.strptime(expiry_datetime, '%d-%m-%Y %H:%M:%S')
                    logger.debug('expiry_datetime=%s', expiry_datetime)
                    
                    now             = datetime.now()
                    if now < expiry_datetime: 
                        logger.debug('auth token is still valid')
                        logger.debug('new args=%s', (user_reference_code, *args))
                        logger.debug('new kwargs=%s', kwargs)
                        
                        
                        
                        if check_duplicated_session:
                            db_client = create_db_client(caller_info="user_auth_token_required")
                            
                            
                            with db_client.context():
                                user_acct           = User.get_by_reference_code(user_reference_code_from_token)
                                signin_device_id    = user_acct.signin_device_id
                            
                            if decrypted_device_id==signin_device_id:
                            
                                return f(*(user_reference_code, *args), **kwargs)
                            else:
                                logger.debug('auth token is not logger valid, due to request device id is not same as latest device id')
                            
                                return create_api_message('Authenticated token is not longer valid', status_code=StatusCode.UNAUTHORIZED, error_code=API_ERR_CODE_DUPLICATED_SESSION)
                        
                        else:
                            return f(*(user_reference_code, *args), **kwargs)
                        
                    else:
                        logger.debug('auth token is not logger valid')
                        
                        #return ("Authenticated token is expired", 401)
                        return create_api_message('Authenticated token is expired', status_code=StatusCode.UNAUTHORIZED, error_code=API_ERR_CODE_EXPIRED_SESSION)
            else:
                #return ("Authenticated token is invalid", 401)    
                return create_api_message('Authenticated token is invalid', status_code=StatusCode.UNAUTHORIZED, error_code=API_ERR_CODE_INVALID_SESSION)
        
        #return ("Authenticated token is required", 401)
        return create_api_message('Authenticated token is required', status_code=StatusCode.UNAUTHORIZED, )
        '''
    return decorated_function

def verify_device(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        user_auth_token     = request.headers.get('x-auth-token')
        user_reference_code = request.headers.get('x-reference-code')
        
        logger.debug('user_auth_token=%s', user_auth_token)
        
        logger.debug('args=%s', args)
        logger.debug('kwargs=%s', kwargs)
        
        if is_not_empty(user_auth_token):
            logger.debug('user_auth_token is not empty, going to decrypt it')
            try:
                auth_details_json   = decrypt_json(user_auth_token)
                logger.debug('auth_details_json=%s', auth_details_json)
                
            except:
                logger.error('Failed due to %s', get_tracelog())
                return create_api_message('Authenticated token is not valid', status_code=StatusCode.UNAUTHORIZED,)
            
            
            
            if auth_details_json:
                expiry_datetime                 = auth_details_json.get('expiry_datetime')
                user_reference_code_from_token  = auth_details_json.get('reference_code')
                decrypted_device_id             = auth_details_json.get('device_id')
                
                if is_not_empty(expiry_datetime) and is_not_empty(user_reference_code_from_token) and user_reference_code_from_token == user_reference_code:
                    
                    expiry_datetime = datetime.strptime(expiry_datetime, '%d-%m-%Y %H:%M:%S')
                    logger.debug('expiry_datetime=%s', expiry_datetime)
                    
                    now             = datetime.now()
                    if now < expiry_datetime: 
                        logger.debug('auth token is still valid')
                        logger.debug('new args=%s', (user_reference_code, *args))
                        logger.debug('new kwargs=%s', kwargs)
                        
                        db_client = create_db_client(caller_info="user_auth_token_required_and_verify_device_token")
                        
                        with db_client.context():
                            user_acct           = User.get_by_reference_code(user_reference_code_from_token)
                            signin_device_id    = user_acct.signin_device_id
                        
                        if decrypted_device_id==signin_device_id:
                        
                            return f(**kwargs)
                        else:
                            logger.debug('auth token is not logger valid, due to ')
                        
                            return create_api_message('Authenticated token is not longer valid', status_code=StatusCode.UNAUTHORIZED,)
                        
                    else:
                        logger.debug('auth token is not logger valid')
                        
                        #return ("Authenticated token is expired", 401)
                        return create_api_message('Authenticated token is expired', status_code=StatusCode.UNAUTHORIZED,)
            else:
                #return ("Authenticated token is invalid", 401)    
                return create_api_message('Authenticated token is invalid', status_code=StatusCode.UNAUTHORIZED,)
        
        #return ("Authenticated token is required", 401)
        return create_api_message('Authenticated token is required', status_code=StatusCode.UNAUTHORIZED,)

    return decorated_function

def device_is_activated(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        auth_token          = request.headers.get('x-auth-token')
        api_access          = False
        
        if is_not_empty(auth_token):
            try:
                auth_details_json = decrypt_json(auth_token)
                logger.debug('auth_details_json=%s', auth_details_json)
                api_access          = auth_details_json.get('api_access', False)
                
                
            except:
                logger.error('Authenticated token is not valid')
                return ("Authenticated token is not valid", 401)
        
        logger.debug('api_access=%s', api_access)
        
        if api_access:
            logger.info('this is api access, thus ignore device activation checking')
            return f(*args, **kwargs)
        else:
            activation_code     = request.headers.get('x-activation-code')
            device_id           = request.headers.get('x-device-id')
            device_type         = request.headers.get('x-device-type')
                
            logger.debug('activation_code=%s', activation_code)
            logger.debug('device_id=%s', device_id)
            logger.debug('device_type=%s', device_type)
            
            if is_not_empty(activation_code) and is_not_empty(device_id):
                device_settings = None
                try:
                    db_client = create_db_client(caller_info="device_is_activated")
                    if 'loyalty' == device_type:
                        with db_client.context():
                            device_settings = LoyaltyDeviceSetting.get_by_activation_code(activation_code)
                    elif 'pos' == device_type:
                        with db_client.context():
                            device_settings = POSSetting.get_by_activation_code(activation_code)
                    
                    if device_settings is not None:
                        if device_settings.testing:
                            logger.info('Activation device id is meant for testing thus, it can be by pass')
                        else:
                            if device_settings.device_id != device_id:
                                logger.info('Access device id is not same as activation device id')
                                return create_api_message('Device is not activated or it is deactivated already', status_code=StatusCode.BAD_REQUEST)
                            
                            logger.info('Access device id is authorized')
                        return f(*args, **kwargs)
                    
                    return create_api_message('Device is not activated or it is deactivated already', status_code=StatusCode.BAD_REQUEST)
                    
                except:
                    logger.error('failed due to %s ', get_tracelog())
                    #return ("Failed to check device authorization", 401)
                    return create_api_message('Failed to check device authorization', status_code=StatusCode.BAD_REQUEST)
                
        
            

    return decorated_function

def device_activated_is_required(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        activation_code     = request.headers.get('x-activation-code')
        device_id           = request.headers.get('x-device-id')
        device_type         = request.headers.get('x-device-type')
            
        logger.debug('activation_code=%s', activation_code)
        logger.debug('device_id=%s', device_id)
        logger.debug('device_type=%s', device_type)
        
        if is_not_empty(activation_code) and is_not_empty(device_id):
            device_settings = None
            try:
                db_client = create_db_client(caller_info="device_is_activated")
                if 'loyalty' == device_type:
                    with db_client.context():
                        device_settings = LoyaltyDeviceSetting.get_by_activation_code(activation_code)
                elif 'pos' == device_type:
                    with db_client.context():
                        device_settings = POSSetting.get_by_activation_code(activation_code)
                
                if device_settings is not None:
                    if device_settings.testing:
                        logger.info('Activation device id is meant for testing thus, it can be by pass')
                    else:
                        if device_settings.device_id != device_id:
                            logger.info('Access device id is not same as activation device id')
                            return create_api_message('Device is not activated or it is deactivated already', status_code=StatusCode.BAD_REQUEST)
                    
                    return f(*(activation_code, *args), *args, **kwargs)
                
                return create_api_message('Device is not activated or it is deactivated already', status_code=StatusCode.BAD_REQUEST)
                
            except:
                logger.error('failed due to %s ', get_tracelog())
                #return ("Failed to check device authorization", 401)
                return create_api_message('Failed to check device is activated', status_code=StatusCode.BAD_REQUEST)
            
        else:  
            return create_api_message('Missing activation code, or device type or device id', status_code=StatusCode.BAD_REQUEST)  

    return decorated_function

def outlet_key_required(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        outlet_key = request.headers.get('x-outlet-key')
            
        logger.debug('outlet_key=%s', outlet_key)
        
        if is_not_empty(outlet_key):
            logger.debug('Going to execute')
            return f(*args, **kwargs)
            
        
        #return ("Outlet Key is required", 401)
        return create_api_message('Missing outlet key', status_code=StatusCode.UNAUTHORIZED)

    return decorated_function

def merchant_key_required(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        acct_id = request.headers.get('x-acct-id')
            
        logger.debug('acct_id=%s', acct_id)
        
        if is_not_empty(acct_id):
            logger.debug('Going to execute')
            return f(*args, **kwargs)
            
        
        #return ("Merchant account id is required", 401)
        return create_api_message('Missing merchant account id', status_code=StatusCode.UNAUTHORIZED)

    return decorated_function

def customer_key_required(f):
    @wraps(f)
    def decorated_function(*args, **kwargs):
        customer_key = request.headers.get('x-customer-key')
            
        logger.debug('customer_key=%s', customer_key)
        
        if is_not_empty(customer_key):
            logger.debug('Going to execute')
            return f(*args, customer_key, **kwargs)
            
        
        #return ("Outlet Key is required", 401)
        return create_api_message('Missing customer key', status_code=StatusCode.UNAUTHORIZED)

    return decorated_function

def elapsed_time_trace(debug=False, trace_key=None):
    def wrapper(fn):
        import time
        def elapsed_time_trace_wrapper(*args, **kwargs):
            start = time.time()
            result      = fn(*args, **kwargs)
            end = time.time()
            elapsed_time = end - start
            trace_name      = trace_key or fn.func_name
            first_argument  = args[0] if args else None
            logger.info('==================== Start Elapsed Time Trace %s(%s) ===========================', trace_name, first_argument)
            logger.info('elapsed time=%s', ("%.2gs" % (elapsed_time)))
            logger.info('================================================================================')
            return result

        return elapsed_time_trace_wrapper
    return wrapper